@@ -6,6 +6,10 @@
 ##  Ralf S. Engelschall <rse@apache.org> for use with 
 ##  Apache's Autoconf-style Interface (APACI).
 ##
+#
+# This script falls under the Apache License.
+# See http://www.apache.org/docs/LICENSE
+
 
 #   input: the modules source file
 modfile=$1
@@ -16,9 +20,9 @@ rm -f $tmpfile
 modname=''
 ext=`echo $modfile | sed 's/^.*\.//'`
 modbase=`echo $modfile | sed 's/\.[^.]*$//'`
-if [ x$ext = x$modfile ]; then ext=o; modbase=$modfile; modfile=$modbase.o; fi
-if [ x$ext = x ] ; then ext=o; modbase=$modfile; fi
-if [ x$ext = xc ] ; then ext=o; fi
+if [ "x$ext" = "x$modfile" ]; then ext=o; modbase=$modfile; modfile=$modbase.o; fi
+if [ "x$ext" = "x" ] ; then ext=o; modbase=$modfile; fi
+if [ "x$ext" = "xc" ] ; then ext=o; fi
 if [ -r $modbase.module ] ; then
     cat $modbase.module >$tmpfile
 else
