@@ -8,11 +8,15 @@
 ##  for pretty printing lists in the --help option of
 ##  Apache's Autoconf-style Interface (APACI)
 ##
+#
+# This script falls under the Apache License.
+# See http://www.apache.org/docs/LICENSE
+
 
 list=`
 IFS=:
 for entry in $*; do
-    if [ ".$entry" != . ]; then
+    if [ "x$entry" != "x" ]; then
         echo $entry
     fi
 done |\
