@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1996-1998 The Apache Group.  All rights reserved.
+ * Copyright (c) 1996-1999 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -58,6 +58,7 @@
 /* Cache and garbage collection routines for Apache proxy */
 
 #include "mod_proxy.h"
+#include "http_conf_globals.h"
 #include "http_log.h"
 #include "http_main.h"
 #include "util_date.h"
@@ -71,33 +72,37 @@
 
 DEF_Explain
 
-#ifndef abs
-#define	abs(c)	((c) >= 0 ? (c) : -(c))
-#endif
-
 struct gc_ent {
     unsigned long int len;
     time_t expire;
     char file[HASH_LEN + 1];
-
 };
 
-static int gcdiff(const void *ap, const void *bp)
-{
-    const struct gc_ent *a = *(const struct gc_ent * const *) ap;
-    const struct gc_ent *b = *(const struct gc_ent * const *) bp;
+/* Poor man's 61 bit arithmetic */
+typedef struct {
+    long lower;	/* lower 30 bits of result */
+    long upper; /* upper 31 bits of result */
+} long61_t;
 
-    if (a->expire > b->expire)
-	return 1;
-    else if (a->expire < b->expire)
-	return -1;
-    else
-	return 0;
-}
+/* FIXME: The block size can be different on a `per file system' base.
+ * This would make automatic detection highly OS specific.
+ * In the GNU fileutils code for du(1), you can see how complicated it can
+ * become to detect the block size. And, with BSD-4.x fragments, it
+ * it even more difficult to get precise results.
+ * As a compromise (and to improve on the incorrect counting of cache
+ * size on byte level, omitting directory sizes entirely, which was
+ * used up to apache-1.3b7) we're rounding to multiples of 512 here.
+ * Your file system may be using larger blocks (I certainly hope so!)
+ * but it will hardly use smaller blocks.
+ * (So this approximation is still closer to reality than the old behavior).
+ * The best solution would be automatic detection, the next best solution
+ * IMHO is a sensible default and the possibility to override it.
+ */
 
-static int curbytes, cachesize, every;
-static unsigned long int curblocks;
-static time_t garbage_now, garbage_expire;
+#define ROUNDUP2BLOCKS(_bytes) (((_bytes)+block_size-1) & ~(block_size-1))
+static long block_size = 512;	/* this must be a power of 2 */
+static long61_t curbytes, cachesize;
+static time_t every, garbage_now, garbage_expire;
 static char *filename;
 static mutex *garbage_mutex = NULL;
 
@@ -114,6 +119,10 @@ int ap_proxy_garbage_init(server_rec *r, pool *p)
 static int sub_garbage_coll(request_rec *r, array_header *files,
 			    const char *cachedir, const char *cachesubdir);
 static void help_proxy_garbage_coll(request_rec *r);
+#if !defined(WIN32) && !defined(MPE) && !defined(OS2)
+static void detached_proxy_garbage_coll(request_rec *r);
+#endif
+
 
 void ap_proxy_garbage_coll(request_rec *r)
 {
@@ -128,7 +137,13 @@ void ap_proxy_garbage_coll(request_rec *r)
 	inside = 1;
     (void) ap_release_mutex(garbage_mutex);
 
+    ap_block_alarms();		/* avoid SIGALRM on big cache cleanup */
+#if !defined(WIN32) && !defined(MPE) && !defined(OS2)
+    detached_proxy_garbage_coll(r);
+#else
     help_proxy_garbage_coll(r);
+#endif
+    ap_unblock_alarms();
 
     (void) ap_acquire_mutex(garbage_mutex);
     inside = 0;
@@ -136,6 +151,119 @@ void ap_proxy_garbage_coll(request_rec *r)
 }
 
 
+static void
+add_long61 (long61_t *accu, long val)
+{
+    /* Add in lower 30 bits */
+    accu->lower += (val & 0x3FFFFFFFL);
+    /* add in upper bits, and carry */
+    accu->upper += (val >> 30) + ((accu->lower & ~0x3FFFFFFFL) != 0L);
+    /* Clear carry */
+    accu->lower &= 0x3FFFFFFFL;
+}
+
+static void
+sub_long61 (long61_t *accu, long val)
+{
+    int carry = (val & 0x3FFFFFFFL) > accu->lower;
+    /* Subtract lower 30 bits */
+    accu->lower = accu->lower - (val & 0x3FFFFFFFL) + ((carry) ? 0x40000000 : 0);
+    /* add in upper bits, and carry */
+    accu->upper -= (val >> 30) + carry;
+}
+
+/* Compare two long61's:
+ * return <0 when left < right
+ * return  0 when left == right
+ * return >0 when left > right
+ */
+static long
+cmp_long61 (long61_t *left, long61_t *right)
+{
+    return (left->upper == right->upper) ? (left->lower - right->lower)
+					 : (left->upper - right->upper);
+}
+
+/* Compare two gc_ent's, sort them by expiration date */
+static int gcdiff(const void *ap, const void *bp)
+{
+    const struct gc_ent *a = (const struct gc_ent * const) ap;
+    const struct gc_ent *b = (const struct gc_ent * const) bp;
+
+    if (a->expire > b->expire)
+	return 1;
+    else if (a->expire < b->expire)
+	return -1;
+    else
+	return 0;
+}
+
+#if !defined(WIN32) && !defined(MPE) && !defined(OS2)
+static void detached_proxy_garbage_coll(request_rec *r)
+{
+    pid_t pid;
+    int status;
+    pid_t pgrp;
+
+    switch (pid = fork()) {
+	case -1:
+	    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+			 "proxy: fork() for cache cleanup failed");
+	    return;
+
+	case 0:	/* Child */
+
+	    /* close all sorts of things, including the socket fd */
+	    ap_cleanup_for_exec();
+
+	    /* Fork twice to disassociate from the child */
+	    switch (pid = fork()) {
+		case -1:
+		    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+			 "proxy: fork(2nd) for cache cleanup failed");
+		    exit(1);
+
+		case 0:	/* Child */
+		    /* The setpgrp() stuff was snarfed from http_main.c */
+#ifndef NO_SETSID
+		    if ((pgrp = setsid()) == -1) {
+			perror("setsid");
+			fprintf(stderr, "%s: setsid failed\n",
+				ap_server_argv0);
+			exit(1);
+		    }
+#elif defined(NEXT) || defined(NEWSOS)
+		    if (setpgrp(0, getpid()) == -1 || (pgrp = getpgrp(0)) == -1) {
+			perror("setpgrp");
+			fprintf(stderr, "%S: setpgrp or getpgrp failed\n",
+				ap_server_argv0);
+			exit(1);
+		    }
+#else
+		    if ((pgrp = setpgrp(getpid(), 0)) == -1) {
+			perror("setpgrp");
+			fprintf(stderr, "%s: setpgrp failed\n",
+				ap_server_argv0);
+			exit(1);
+		    }
+#endif
+		    help_proxy_garbage_coll(r);
+		    exit(0);
+
+		default:    /* Father */
+		    /* After grandson has been forked off, */
+		    /* there's nothing else to do. */
+		    exit(0);		    
+	    }
+	default:
+	    /* Wait until grandson has been forked off */
+	    /* (without wait we'd leave a zombie) */
+	    waitpid(pid, &status, 0);
+	    return;
+    }
+}
+#endif /* ndef WIN32 */
+
 static void help_proxy_garbage_coll(request_rec *r)
 {
     const char *cachedir;
@@ -145,17 +273,24 @@ static void help_proxy_garbage_coll(request_rec *r)
     const struct cache_conf *conf = &pconf->cache;
     array_header *files;
     struct stat buf;
-    struct gc_ent *fent, **elts;
+    struct gc_ent *fent;
     int i, timefd;
-    static time_t lastcheck = BAD_DATE;		/* static data!!! */
+    static time_t lastcheck = BAD_DATE;		/* static (per-process) data!!! */
 
     cachedir = conf->root;
-    cachesize = conf->space;
+    /* configured size is given in kB. Make it bytes, convert to long61_t: */
+    cachesize.lower = cachesize.upper = 0;
+    add_long61(&cachesize, conf->space << 10);
     every = conf->gcinterval;
 
     if (cachedir == NULL || every == -1)
 	return;
     garbage_now = time(NULL);
+    /* Usually, the modification time of <cachedir>/.time can only increase.
+     * Thus, even with several child processes having their own copy of
+     * lastcheck, if time(NULL) still < lastcheck then it's not time
+     * for GC yet.
+     */
     if (garbage_now != -1 && lastcheck != BAD_DATE && garbage_now < lastcheck + every)
 	return;
 
@@ -176,7 +311,7 @@ static void help_proxy_garbage_coll(request_rec *r)
 		ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
 			     "proxy: creat(%s)", filename);
 	    else
-		lastcheck = abs(garbage_now);	/* someone else got in there */
+		lastcheck = garbage_now;	/* someone else got in there */
 	    ap_unblock_alarms();
 	    return;
 	}
@@ -192,22 +327,24 @@ static void help_proxy_garbage_coll(request_rec *r)
 	    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
 			 "proxy: utimes(%s)", filename);
     }
-    files = ap_make_array(r->pool, 100, sizeof(struct gc_ent *));
-    curblocks = 0;
-    curbytes = 0;
+    files = ap_make_array(r->pool, 100, sizeof(struct gc_ent));
+    curbytes.upper = curbytes.lower = 0L;
 
     sub_garbage_coll(r, files, cachedir, "/");
 
-    if (curblocks < cachesize || curblocks + curbytes <= cachesize) {
+    if (cmp_long61(&curbytes, &cachesize) < 0L) {
+	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, r->server,
+			 "proxy GC: Cache is %ld%% full (nothing deleted)",
+			 (long)(((curbytes.upper<<20)|(curbytes.lower>>10))*100/conf->space));
 	ap_unblock_alarms();
 	return;
     }
 
-    qsort(files->elts, files->nelts, sizeof(struct gc_ent *), gcdiff);
+    /* sort the files we found by expiration date */
+    qsort(files->elts, files->nelts, sizeof(struct gc_ent), gcdiff);
 
-    elts = (struct gc_ent **) files->elts;
     for (i = 0; i < files->nelts; i++) {
-	fent = elts[i];
+	fent = &((struct gc_ent *) files->elts)[i];
 	sprintf(filename, "%s%s", cachedir, fent->file);
 	Explain3("GC Unlinking %s (expiry %ld, garbage_now %ld)", filename, fent->expire, garbage_now);
 #if TESTING
@@ -221,16 +358,15 @@ static void help_proxy_garbage_coll(request_rec *r)
 	else
 #endif
 	{
-	    curblocks -= fent->len >> 10;
-	    curbytes -= fent->len & 0x3FF;
-	    if (curbytes < 0) {
-		curbytes += 1024;
-		curblocks--;
-	    }
-	    if (curblocks < cachesize || curblocks + curbytes <= cachesize)
+	    sub_long61(&curbytes, ROUNDUP2BLOCKS(fent->len));
+	    if (cmp_long61(&curbytes, &cachesize) < 0)
 		break;
 	}
     }
+
+    ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, r->server,
+			 "proxy GC: Cache is %ld%% full (%d deleted)",
+			 (long)(((curbytes.upper<<20)|(curbytes.lower>>10))*100/conf->space), i);
     ap_unblock_alarms();
 }
 
@@ -291,7 +427,7 @@ static int sub_garbage_coll(request_rec *r, array_header *files,
 	/*      if (strlen(ent->d_name) != HASH_LEN) continue; */
 
 /* under OS/2 use dirent's d_attr to identify a diretory */
-#ifdef __EMX__
+#ifdef OS2
 /* is it a directory? */
 	if (ent->d_attr & A_DIR) {
 	    char newcachedir[HUGE_STRING_LEN];
@@ -327,7 +463,7 @@ static int sub_garbage_coll(request_rec *r, array_header *files,
 	}
 
 /* In OS/2 this has already been done above */
-#ifndef __EMX__
+#ifndef OS2
 	if (S_ISDIR(buf.st_mode)) {
 	    char newcachedir[HUGE_STRING_LEN];
 	    close(fd);
@@ -342,6 +478,9 @@ static int sub_garbage_coll(request_rec *r, array_header *files,
 		rmdir(newcachedir);
 #endif
 		--nfiles;
+	    } else {
+		/* Directory is not empty. Account for its size: */
+		add_long61(&curbytes, ROUNDUP2BLOCKS(buf.st_size));
 	    }
 	    continue;
 	}
@@ -378,22 +517,14 @@ static int sub_garbage_coll(request_rec *r, array_header *files,
  * file.
  *
  */
-	/* FIXME: We should make the array an array of gc_ents, not gc_ent *s
-	 */
-	fent = ap_palloc(r->pool, sizeof(struct gc_ent));
+	fent = (struct gc_ent *) ap_push_array(files);
 	fent->len = buf.st_size;
 	fent->expire = garbage_expire;
 	strcpy(fent->file, cachesubdir);
 	strcat(fent->file, ent->d_name);
-	*(struct gc_ent **) ap_push_array(files) = fent;
 
 /* accumulate in blocks, to cope with directories > 4Gb */
-	curblocks += buf.st_size >> 10;		/* Kbytes */
-	curbytes += buf.st_size & 0x3FF;
-	if (curbytes >= 1024) {
-	    curbytes -= 1024;
-	    curblocks++;
-	}
+	add_long61(&curbytes, ROUNDUP2BLOCKS(buf.st_size));
     }
 
     closedir(dir);
@@ -408,7 +539,7 @@ static int sub_garbage_coll(request_rec *r, array_header *files,
  *         0 on failure (bad file or wrong URL)
  *        -1 on UNIX error
  */
-static int rdcache(pool *p, BUFF *cachefp, struct cache_req *c)
+static int rdcache(request_rec *r, BUFF *cachefp, cache_req *c)
 {
     char urlbuff[1034], *strp;
     int len;
@@ -417,7 +548,7 @@ static int rdcache(pool *p, BUFF *cachefp, struct cache_req *c)
  * date SP lastmod SP expire SP count SP content-length CRLF
  * dates are stored as hex seconds since 1970
  */
-    len = ap_bgets(urlbuff, 1034, cachefp);
+    len = ap_bgets(urlbuff, sizeof urlbuff, cachefp);
     if (len == -1)
 	return -1;
     if (len == 0 || urlbuff[len - 1] != '\n')
@@ -435,7 +566,7 @@ static int rdcache(pool *p, BUFF *cachefp, struct cache_req *c)
     c->len = ap_proxy_hex2sec(urlbuff + 36);
 
 /* check that we have the same URL */
-    len = ap_bgets(urlbuff, 1034, cachefp);
+    len = ap_bgets(urlbuff, sizeof urlbuff, cachefp);
     if (len == -1)
 	return -1;
     if (len == 0 || strncmp(urlbuff, "X-URL: ", 7) != 0 ||
@@ -446,29 +577,26 @@ static int rdcache(pool *p, BUFF *cachefp, struct cache_req *c)
 	return 0;
 
 /* What follows is the message */
-    len = ap_bgets(urlbuff, 1034, cachefp);
+    len = ap_bgets(urlbuff, sizeof urlbuff, cachefp);
     if (len == -1)
 	return -1;
     if (len == 0 || urlbuff[len - 1] != '\n')
 	return 0;
     urlbuff[--len] = '\0';
 
-    c->resp_line = ap_pstrdup(p, urlbuff);
+    c->resp_line = ap_pstrdup(r->pool, urlbuff);
     strp = strchr(urlbuff, ' ');
     if (strp == NULL)
 	return 0;
 
     c->status = atoi(strp);
-    c->hdrs = ap_proxy_read_headers(p, urlbuff, 1034, cachefp);
+    c->hdrs = ap_proxy_read_headers(r, urlbuff, sizeof urlbuff, cachefp);
     if (c->hdrs == NULL)
 	return -1;
     if (c->len != -1) {		/* add a content-length header */
-	struct hdr_entry *q;
-	q = ap_proxy_get_header(c->hdrs, "Content-Length");
-	if (q == NULL) {
-	    strp = ap_palloc(p, 15);
-	    ap_snprintf(strp, 15, "%u", c->len);
-	    ap_proxy_add_header(c->hdrs, "Content-Length", strp, HDR_REP);
+	if (ap_table_get(c->hdrs, "Content-Length") == NULL) {
+	    ap_table_set(c->hdrs, "Content-Length",
+			 ap_psprintf(r->pool, "%lu", (unsigned long)c->len));
 	}
     }
     return 1;
@@ -490,10 +618,11 @@ static int rdcache(pool *p, BUFF *cachefp, struct cache_req *c)
  *            last modified date to request
  */
 int ap_proxy_cache_check(request_rec *r, char *url, struct cache_conf *conf,
-		      struct cache_req **cr)
+		      cache_req **cr)
 {
-    char hashfile[66], *imstr, *pragma, *auth;
-    struct cache_req *c;
+    char hashfile[66];
+    const char *imstr, *pragma, *auth;
+    cache_req *c;
     time_t now;
     BUFF *cachefp;
     int cfd, i;
@@ -502,7 +631,7 @@ int ap_proxy_cache_check(request_rec *r, char *url, struct cache_conf *conf,
     proxy_server_conf *pconf =
     (proxy_server_conf *) ap_get_module_config(sconf, &proxy_module);
 
-    c = ap_pcalloc(r->pool, sizeof(struct cache_req));
+    c = ap_pcalloc(r->pool, sizeof(cache_req));
     *cr = c;
     c->req = r;
     c->url = ap_pstrdup(r->pool, url);
@@ -542,7 +671,7 @@ int ap_proxy_cache_check(request_rec *r, char *url, struct cache_conf *conf,
 	    ap_bpushfd(cachefp, cfd, cfd);
 	}
 	else if (errno != ENOENT)
-	    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 			 "proxy: error opening cache file %s",
 			 c->filename);
 #ifdef EXPLAIN
@@ -552,21 +681,24 @@ int ap_proxy_cache_check(request_rec *r, char *url, struct cache_conf *conf,
     }
 
     if (cachefp != NULL) {
-	i = rdcache(r->pool, cachefp, c);
+	i = rdcache(r, cachefp, c);
 	if (i == -1)
-	    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 			 "proxy: error reading cache file %s", 
 			 c->filename);
 	else if (i == 0)
-	    ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, r->server,
+	    ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, r,
 			 "proxy: bad (short?) cache file: %s", c->filename);
 	if (i != 1) {
 	    ap_pclosef(r->pool, cachefp->fd);
 	    cachefp = NULL;
 	}
     }
+/* fixed?  in this case, we want to get the headers from the remote server
+   it will be handled later if we don't do this (I hope ;-)
     if (cachefp == NULL)
-	c->hdrs = ap_make_array(r->pool, 2, sizeof(struct hdr_entry));
+	c->hdrs = ap_make_table(r->pool, 20);
+*/
     /* FIXME: Shouldn't we check the URL somewhere? */
     now = time(NULL);
 /* Ok, have we got some un-expired data? */
@@ -577,20 +709,16 @@ int ap_proxy_cache_check(request_rec *r, char *url, struct cache_conf *conf,
 /* has the cached file changed since this request? */
 	    if (c->date == BAD_DATE || c->date > c->ims) {
 /* No, but these header values may have changed, so we send them with the
- * 304 response
+ * 304 HTTP_NOT_MODIFIED response
  */
-		/* CHECKME: surely this was wrong? (Ben)
-		   p = table_get(r->headers_in, "Expires");
-		 */
-		struct hdr_entry *q;
+		const char *q;
 
-		q = ap_proxy_get_header(c->hdrs, "Expires");
-		if (q != NULL && q->value != NULL)
-		    ap_table_set(r->headers_out, "Expires", q->value);
+		if ((q = ap_table_get(c->hdrs, "Expires")) != NULL)
+		    ap_table_set(r->headers_out, "Expires", q);
 	    }
 	    ap_pclosef(r->pool, cachefp->fd);
 	    Explain0("Use local copy, cached file hasn't changed");
-	    return USE_LOCAL_COPY;
+	    return HTTP_NOT_MODIFIED;
 	}
 
 /* Ok, has been modified */
@@ -605,7 +733,7 @@ int ap_proxy_cache_check(request_rec *r, char *url, struct cache_conf *conf,
 	ap_bsetopt(r->connection->client, BO_BYTECT, &zero);
 	r->sent_bodyct = 1;
 	if (!r->header_only)
-	    ap_proxy_send_fb(cachefp, r, NULL, NULL);
+	    ap_proxy_send_fb(cachefp, r, NULL);
 	ap_pclosef(r->pool, cachefp->fd);
 	return OK;
     }
@@ -620,13 +748,11 @@ int ap_proxy_cache_check(request_rec *r, char *url, struct cache_conf *conf,
  * from the cache
  */
 	if (c->ims == BAD_DATE || c->ims < c->lmod) {
-	    struct hdr_entry *q;
+	    const char *q;
 
-	    q = ap_proxy_get_header(c->hdrs, "Last-Modified");
-
-	    if (q != NULL && q->value != NULL)
+	    if ((q = ap_table_get(c->hdrs, "Last-Modified")) != NULL)
 		ap_table_set(r->headers_in, "If-Modified-Since",
-			  (char *) q->value);
+			  (char *) q);
 	}
     }
     c->fp = cachefp;
@@ -648,7 +774,7 @@ int ap_proxy_cache_check(request_rec *r, char *url, struct cache_conf *conf,
  *  from the cache, maybe updating the header line
  *  otherwise, delete the old cached file and open a new temporary file
  */
-int ap_proxy_cache_update(struct cache_req *c, array_header *resp_hdrs,
+int ap_proxy_cache_update(cache_req *c, table *resp_hdrs,
 		       const int is_HTTP1, int nocache)
 {
 #ifdef ULTRIX_BRAIN_DEATH
@@ -657,7 +783,7 @@ int ap_proxy_cache_update(struct cache_req *c, array_header *resp_hdrs,
     request_rec *r = c->req;
     char *p;
     int i;
-    struct hdr_entry *expire, *dates, *lmods, *clen;
+    const char *expire, *lmods, *dates, *clen;
     time_t expc, date, lmod, now;
     char buff[46];
     void *sconf = r->server->module_config;
@@ -671,21 +797,20 @@ int ap_proxy_cache_update(struct cache_req *c, array_header *resp_hdrs,
 /* read expiry date; if a bad date, then leave it so the client can
  * read it
  */
-    expire = ap_proxy_get_header(resp_hdrs, "Expires");
+    expire = ap_table_get(resp_hdrs, "Expires");
     if (expire != NULL)
-	expc = ap_parseHTTPdate(expire->value);
+	expc = ap_parseHTTPdate(expire);
     else
 	expc = BAD_DATE;
 
 /*
  * read the last-modified date; if the date is bad, then delete it
  */
-    lmods = ap_proxy_get_header(resp_hdrs, "Last-Modified");
+    lmods = ap_table_get(resp_hdrs, "Last-Modified");
     if (lmods != NULL) {
-	lmod = ap_parseHTTPdate(lmods->value);
+	lmod = ap_parseHTTPdate(lmods);
 	if (lmod == BAD_DATE) {
 /* kill last modified date */
-	    lmods->value = NULL;
 	    lmods = NULL;
 	}
     }
@@ -695,16 +820,18 @@ int ap_proxy_cache_update(struct cache_req *c, array_header *resp_hdrs,
 /*
  * what responses should we not cache?
  * Unknown status responses and those known to be uncacheable
- * 304 response when we have no valid cache file, or
- * 200 response from HTTP/1.0 and up without a Last-Modified header, or
+ * 304 HTTP_NOT_MODIFIED response when we have no valid cache file, or
+ * 200 HTTP_OK response from HTTP/1.0 and up without a Last-Modified header, or
  * HEAD requests, or
  * requests with an Authorization header, or
  * protocol requests nocache (e.g. ftp with user/password)
  */
-    if ((r->status != 200 && r->status != 301 && r->status != 304) ||
+/* @@@ XXX FIXME: is the test "r->status != HTTP_MOVED_PERMANENTLY" correct?
+ * or shouldn't it be "ap_is_HTTP_REDIRECT(r->status)" ? -MnKr */
+    if ((r->status != HTTP_OK && r->status != HTTP_MOVED_PERMANENTLY && r->status != HTTP_NOT_MODIFIED) ||
 	(expire != NULL && expc == BAD_DATE) ||
-	(r->status == 304 && c->fp == NULL) ||
-	(r->status == 200 && lmods == NULL && is_HTTP1) ||
+	(r->status == HTTP_NOT_MODIFIED && (c == NULL || c->fp == NULL)) ||
+	(r->status == HTTP_OK && lmods == NULL && is_HTTP1) ||
 	r->header_only ||
 	ap_table_get(r->headers_in, "Authorization") != NULL ||
 	nocache) {
@@ -715,7 +842,8 @@ int ap_proxy_cache_update(struct cache_req *c, array_header *resp_hdrs,
 	    c->fp = NULL;
 	}
 /* delete the previously cached file */
-	unlink(c->filename);
+        if (c->filename)
+            unlink(c->filename);
 	return DECLINED;	/* send data to client but not cache */
     }
 
@@ -723,9 +851,9 @@ int ap_proxy_cache_update(struct cache_req *c, array_header *resp_hdrs,
 /*
  * Read the date. Generate one if one is not supplied
  */
-    dates = ap_proxy_get_header(resp_hdrs, "Date");
+    dates = ap_table_get(resp_hdrs, "Date");
     if (dates != NULL)
-	date = ap_parseHTTPdate(dates->value);
+	date = ap_parseHTTPdate(dates);
     else
 	date = BAD_DATE;
 
@@ -735,9 +863,9 @@ int ap_proxy_cache_update(struct cache_req *c, array_header *resp_hdrs,
 /* no date header! */
 /* add one; N.B. use the time _now_ rather than when we were checking the cache
  */
-	date = abs(now);
-	p = ap_gm_timestr_822(r->pool, now);
-	dates = ap_proxy_add_header(resp_hdrs, "Date", p, HDR_REP);
+	date = now;
+	dates = ap_gm_timestr_822(r->pool, now);
+	ap_table_set(resp_hdrs, "Date", dates);
 	Explain0("Added date header");
     }
 
@@ -746,7 +874,7 @@ int ap_proxy_cache_update(struct cache_req *c, array_header *resp_hdrs,
 /* if its in the future, then replace by date */
     {
 	lmod = date;
-	lmods->value = dates->value;
+	lmods = dates;
 	Explain0("Last modified is in the future, replacing with now");
     }
 /* if the response did not contain the header, then use the cached version */
@@ -757,9 +885,9 @@ int ap_proxy_cache_update(struct cache_req *c, array_header *resp_hdrs,
 
 /* we now need to calculate the expire data for the object. */
     if (expire == NULL && c->fp != NULL) {	/* no expiry data sent in response */
-	expire = ap_proxy_get_header(c->hdrs, "Expires");
+	expire = ap_table_get(c->hdrs, "Expires");
 	if (expire != NULL)
-	    expc = ap_parseHTTPdate(expire->value);
+	    expc = ap_parseHTTPdate(expire);
     }
 /* so we now have the expiry date */
 /* if no expiry date then
@@ -775,19 +903,19 @@ int ap_proxy_cache_update(struct cache_req *c, array_header *resp_hdrs,
 	    double maxex = conf->cache.maxexpire;
 	    if (x > maxex)
 		x = maxex;
-	    expc = abs(now) + (int) x;
+	    expc = now + (int) x;
 	}
 	else
-	    expc = abs(now) + conf->cache.defaultexpire;
+	    expc = now + conf->cache.defaultexpire;
 	Explain1("Expiry date calculated %ld", expc);
     }
 
 /* get the content-length header */
-    clen = ap_proxy_get_header(resp_hdrs, "Content-Length");
+    clen = ap_table_get(resp_hdrs, "Content-Length");
     if (clen == NULL)
 	c->len = -1;
     else
-	c->len = atoi(clen->value);
+	c->len = atoi(clen);
 
     ap_proxy_sec2hex(date, buff);
     buff[8] = ' ';
@@ -802,25 +930,25 @@ int ap_proxy_cache_update(struct cache_req *c, array_header *resp_hdrs,
     buff[45] = '\0';
 
 /* if file not modified */
-    if (r->status == 304) {
+    if (r->status == HTTP_NOT_MODIFIED) {
 	if (c->ims != BAD_DATE && lmod != BAD_DATE && lmod <= c->ims) {
 /* set any changed headers somehow */
 /* update dates and version, but not content-length */
 	    if (lmod != c->lmod || expc != c->expire || date != c->date) {
 		off_t curpos = lseek(c->fp->fd, 0, SEEK_SET);
 		if (curpos == -1)
-		    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+		    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 				 "proxy: error seeking on cache file %s",
 				 c->filename);
 		else if (write(c->fp->fd, buff, 35) == -1)
-		    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+		    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 				 "proxy: error updating cache file %s",
 				 c->filename);
 	    }
 	    ap_pclosef(r->pool, c->fp->fd);
 	    Explain0("Remote document not modified, use local copy");
 	    /* CHECKME: Is this right? Shouldn't we check IMS again here? */
-	    return USE_LOCAL_COPY;
+	    return HTTP_NOT_MODIFIED;
 	}
 	else {
 /* return the whole document */
@@ -835,18 +963,18 @@ int ap_proxy_cache_update(struct cache_req *c, array_header *resp_hdrs,
 	    ap_bsetopt(r->connection->client, BO_BYTECT, &zero);
 	    r->sent_bodyct = 1;
 	    if (!r->header_only)
-		ap_proxy_send_fb(c->fp, r, NULL, NULL);
+		ap_proxy_send_fb(c->fp, r, NULL);
 /* set any changed headers somehow */
 /* update dates and version, but not content-length */
 	    if (lmod != c->lmod || expc != c->expire || date != c->date) {
 		off_t curpos = lseek(c->fp->fd, 0, SEEK_SET);
 
 		if (curpos == -1)
-		    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+		    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 				 "proxy: error seeking on cache file %s",
 				 c->filename);
 		else if (write(c->fp->fd, buff, 35) == -1)
-		    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+		    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 				 "proxy: error updating cache file %s",
 				 c->filename);
 	    }
@@ -879,7 +1007,7 @@ int ap_proxy_cache_update(struct cache_req *c, array_header *resp_hdrs,
 
     i = open(c->tempfile, O_WRONLY | O_CREAT | O_EXCL | O_BINARY, 0622);
     if (i == -1) {
-	ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 		     "proxy: error creating cache file %s",
 		     c->tempfile);
 	return DECLINED;
@@ -889,7 +1017,7 @@ int ap_proxy_cache_update(struct cache_req *c, array_header *resp_hdrs,
     ap_bpushfd(c->fp, -1, i);
 
     if (ap_bvputs(c->fp, buff, "X-URL: ", c->url, "\n", NULL) == -1) {
-	ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 		     "proxy: error writing cache file(%s)", c->tempfile);
 	ap_pclosef(r->pool, c->fp->fd);
 	unlink(c->tempfile);
@@ -898,15 +1026,20 @@ int ap_proxy_cache_update(struct cache_req *c, array_header *resp_hdrs,
     return DECLINED;
 }
 
-void ap_proxy_cache_tidy(struct cache_req *c)
+void ap_proxy_cache_tidy(cache_req *c)
 {
-    server_rec *s = c->req->server;
+    server_rec *s;
     long int bc;
 
-    if (c->fp == NULL)
+    if (c == NULL || c->fp == NULL)
 	return;
 
+    s = c->req->server;
+
+/* don't care how much was sent, but rather how much was written to cache
     ap_bgetopt(c->req->connection->client, BO_BYTECT, &bc);
+ */
+    bc = c->written;
 
     if (c->len != -1) {
 /* file lengths don't match; don't cache it */
@@ -916,11 +1049,13 @@ void ap_proxy_cache_tidy(struct cache_req *c)
 	    return;
 	}
     }
+/* don't care if aborted, cache it if fully retrieved from host!
     else if (c->req->connection->aborted) {
-	ap_pclosef(c->req->pool, c->fp->fd);	/* no need to flush */
+	ap_pclosef(c->req->pool, c->fp->fd);	/ no need to flush /
 	unlink(c->tempfile);
 	return;
     }
+*/
     else {
 /* update content-length of file */
 	char buff[9];
@@ -980,7 +1115,7 @@ void ap_proxy_cache_tidy(struct cache_req *c)
 	    *p = '/';
 	    ++p;
 	}
-#if defined(__EMX__) || defined(WIN32)
+#if defined(OS2) || defined(WIN32)
 	/* Under OS/2 use rename. */
 	if (rename(c->tempfile, c->filename) == -1)
 	    ap_log_error(APLOG_MARK, APLOG_ERR, s,
