@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1996-1998 The Apache Group.  All rights reserved.
+ * Copyright (c) 1996-1999 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -61,6 +61,8 @@
 #include "http_main.h"
 #include "http_log.h"
 
+#define AUTODETECT_PWD
+
 DEF_Explain
 
 /*
@@ -75,7 +77,7 @@ static int decodeenc(char *x)
     for (i = 0, j = 0; x[i] != '\0'; i++, j++) {
 /* decode it if not already done */
 	ch = x[i];
-	if (ch == '%' && isxdigit(x[i + 1]) && isxdigit(x[i + 2])) {
+	if (ch == '%' && ap_isxdigit(x[i + 1]) && ap_isxdigit(x[i + 2])) {
 	    ch = ap_proxy_hex2c(&x[i + 1]);
 	    i += 2;
 	}
@@ -95,7 +97,7 @@ static int ftp_check_string(const char *x)
 
     for (i = 0; x[i] != '\0'; i++) {
 	ch = x[i];
-	if (ch == '%' && isxdigit(x[i + 1]) && isxdigit(x[i + 2])) {
+	if (ch == '%' && ap_isxdigit(x[i + 1]) && ap_isxdigit(x[i + 2])) {
 	    ch = ap_proxy_hex2c(&x[i + 1]);
 	    i += 2;
 	}
@@ -195,8 +197,8 @@ static int ftp_getrc(BUFF *f)
     if (len == -1)
 	return -1;
 /* check format */
-    if (len < 5 || !isdigit(linebuff[0]) || !isdigit(linebuff[1]) ||
-	!isdigit(linebuff[2]) || (linebuff[3] != ' ' && linebuff[3] != '-'))
+    if (len < 5 || !ap_isdigit(linebuff[0]) || !ap_isdigit(linebuff[1]) ||
+	!ap_isdigit(linebuff[2]) || (linebuff[3] != ' ' && linebuff[3] != '-'))
 	status = 0;
     else
 	status = 100 * linebuff[0] + 10 * linebuff[1] + linebuff[2] - 111 * '0';
@@ -236,8 +238,8 @@ static int ftp_getrc_msg(BUFF *f, char *msgbuf, int msglen)
     len = ap_bgets(linebuff, sizeof linebuff, f);
     if (len == -1)
 	return -1;
-    if (len < 5 || !isdigit(linebuff[0]) || !isdigit(linebuff[1]) ||
-	!isdigit(linebuff[2]) || (linebuff[3] != ' ' && linebuff[3] != '-'))
+    if (len < 5 || !ap_isdigit(linebuff[0]) || !ap_isdigit(linebuff[1]) ||
+	!ap_isdigit(linebuff[2]) || (linebuff[3] != ' ' && linebuff[3] != '-'))
 	status = 0;
     else
 	status = 100 * linebuff[0] + 10 * linebuff[1] + linebuff[2] - 111 * '0';
@@ -263,65 +265,38 @@ static int ftp_getrc_msg(BUFF *f, char *msgbuf, int msglen)
     return status;
 }
 
-static long int send_dir(BUFF *f, request_rec *r, BUFF *f2, struct cache_req *c, char *url)
+static long int send_dir(BUFF *f, request_rec *r, cache_req *c, char *cwd)
 {
     char buf[IOBUFSIZE];
     char buf2[IOBUFSIZE];
     char *filename;
-    char *tempurl;
     int searchidx = 0;
     char *searchptr = NULL;
     int firstfile = 1;
-    char urlptr[HUGE_STRING_LEN];
     unsigned long total_bytes_sent = 0;
     register int n, o, w;
-    int hostlen;
     conn_rec *con = r->connection;
-    char *dir, *path, *reldir, *site, *psite;
-    const char *sig;
+    char *dir, *path, *reldir, *site;
 
-    tempurl = ap_pstrdup(r->pool, url);
-
-    (void)decodeenc(tempurl);
-
-    /* Determine length of "scheme://site" prefix */
-    for (hostlen=0; tempurl[hostlen]!='/'; ++hostlen)
-	continue;
-    if (tempurl[hostlen] == '/' && tempurl[hostlen+1] == '/') {
-	for (hostlen+=2; tempurl[hostlen]!='/' && tempurl[hostlen]!='?'; ++hostlen)
-	    continue;
-    } else {
-	hostlen = 0;
-    }
-
-    /* Save "scheme://site" prefix */
-    site = psite = ap_pstrndup(r->pool, tempurl, hostlen);
-
-    if ((n = strcspn(tempurl, "@")) != strlen(tempurl) && n < hostlen) {    /* hide user/passwd */
-	memmove(tempurl + (n - 5), tempurl, 6);
-	tempurl += n - 5;	/* leave room for ftp:// */
-	hostlen -= (n-5);
-
-	/* Save "scheme://site" prefix without user/password */
-	site = ap_pstrndup(r->pool, tempurl, hostlen);
-    }
+    /* Save "scheme://site" prefix without password */
+    site = ap_unparse_uri_components(r->pool, &r->parsed_uri, UNP_OMITPASSWORD|UNP_OMITPATHINFO);
+    /* ... and path without query args */
+    path = ap_unparse_uri_components(r->pool, &r->parsed_uri, UNP_OMITSITEPART|UNP_OMITQUERY);
+    (void)decodeenc(path);
 
     /* Copy path, strip (all except the last) trailing slashes */
-    path = dir = ap_pstrcat(r->pool, tempurl+hostlen, "/", NULL);
+    path = dir = ap_pstrcat(r->pool, path, "/", NULL);
     while ((n = strlen(path)) > 1 && path[n-1] == '/' && path[n-2] == '/')
 	path[n-1] = '\0';
 
     /* print "ftp://host/" */
-    ap_snprintf(buf, sizeof(buf), "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">\n"
-		"<HTML><HEAD><TITLE>%s</TITLE>\n"
+    n = ap_snprintf(buf, sizeof(buf), "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">\n"
+		"<HTML><HEAD><TITLE>%s%s</TITLE>\n"
 		"<BASE HREF=\"%s%s\"></HEAD>\n"
 		"<BODY><H2>Directory of "
 		"<A HREF=\"/\">%s</A>/",
-		tempurl, psite, path, site);
-    ap_bputs(buf, con->client);
-    if (f2 != NULL)
-	ap_bputs(buf, f2);
-    total_bytes_sent += strlen(buf);
+		site, path, site, path, site);
+    total_bytes_sent += ap_proxy_bputs2(buf, con->client, c);
 
     while ((dir = strchr(dir+1, '/')) != NULL)
     {
@@ -332,31 +307,23 @@ static long int send_dir(BUFF *f, request_rec *r, BUFF *f2, struct cache_req *c,
 	    ++reldir;
 	/* print "path/" component */
 	ap_snprintf(buf, sizeof(buf), "<A HREF=\"/%s/\">%s</A>/", path+1, reldir);
-	ap_bputs(buf, con->client);
-    if (f2 != NULL)
-	    ap_bputs(buf, f2);
-	total_bytes_sent += strlen(buf);
+	total_bytes_sent += ap_proxy_bputs2(buf, con->client, c);
 	*dir = '/';
     }
-    ap_snprintf(buf, sizeof(buf), "</H2>\n<HR><PRE>");
-    ap_bputs(buf, con->client);
-    if (f2 != NULL)
-	    ap_bputs(buf, f2);
-    total_bytes_sent += strlen(buf);
-
-    for (hostlen=0; url[hostlen]!='/'; ++hostlen)
-	continue;
-    if (url[hostlen] == '/' && url[hostlen+1] == '/') {
-	for (hostlen+=2; url[hostlen]!='/' && url[hostlen]!='?'; ++hostlen)
-	    continue;
-    } else
-	hostlen = 0;
+    /* If the caller has determined the current directory, and it differs */
+    /* from what the client requested, then show the real name */
+    if (cwd == NULL || strncmp (cwd, path, strlen(cwd)) == 0) {
+	ap_snprintf(buf, sizeof(buf), "</H2>\n<HR><PRE>");
+    } else {
+	ap_snprintf(buf, sizeof(buf), "</H2>\n(%s)\n<HR><PRE>", cwd);
+    }
+    total_bytes_sent += ap_proxy_bputs2(buf, con->client, c);
 
     while (!con->aborted) {
 	n = ap_bgets(buf, sizeof buf, f);
 	if (n == -1) {		/* input error */
-	    if (f2 != NULL)
-		f2 = ap_proxy_cache_error(c);
+	    if (c != NULL)
+		c = ap_proxy_cache_error(c);
 	    break;
 	}
 	if (n == 0)
@@ -364,20 +331,19 @@ static long int send_dir(BUFF *f, request_rec *r, BUFF *f2, struct cache_req *c,
 	if (buf[0] == 'l' && (filename=strstr(buf, " -> ")) != NULL) {
 	    char *link_ptr = filename;
 
-	    do
+	    do {
 		filename--;
-	    while (filename[0] != ' ');
-	    *(filename++) = 0;
-	    *(link_ptr++) = 0;
+	    } while (filename[0] != ' ');
+	    *(filename++) = '\0';
+	    *(link_ptr++) = '\0';
 	    if ((n = strlen(link_ptr)) > 1 && link_ptr[n - 1] == '\n')
 	      link_ptr[n - 1] = '\0';
-	    ap_snprintf(urlptr, sizeof(urlptr), "%s%s%s", url+hostlen, (url[strlen(url) - 1] == '/' ? "" : "/"), filename);
 	    ap_snprintf(buf2, sizeof(buf2), "%s <A HREF=\"%s\">%s %s</A>\n", buf, filename, filename, link_ptr);
 	    ap_cpystrn(buf, buf2, sizeof(buf));
 	    n = strlen(buf);
 	}
-	else if (buf[0] == 'd' || buf[0] == '-' || buf[0] == 'l' || isdigit(buf[0])) {
-	    if (isdigit(buf[0])) {	/* handle DOS dir */
+	else if (buf[0] == 'd' || buf[0] == '-' || buf[0] == 'l' || ap_isdigit(buf[0])) {
+	    if (ap_isdigit(buf[0])) {	/* handle DOS dir */
 		searchptr = strchr(buf, '<');
 		if (searchptr != NULL)
 		    *searchptr = '[';
@@ -416,9 +382,8 @@ static long int send_dir(BUFF *f, request_rec *r, BUFF *f2, struct cache_req *c,
 	o = 0;
 	total_bytes_sent += n;
 
-	if (f2 != NULL)
-	    if (ap_bwrite(f2, buf, n) != n)
-		f2 = ap_proxy_cache_error(c);
+	if (c != NULL && c->fp && ap_bwrite(c->fp, buf, n) != n)
+	    c = ap_proxy_cache_error(c);
 
 	while (n && !r->connection->aborted) {
 	    w = ap_bwrite(con->client, &buf[o], n);
@@ -429,52 +394,71 @@ static long int send_dir(BUFF *f, request_rec *r, BUFF *f2, struct cache_req *c,
 	    o += w;
 	}
     }
-    site = "</PRE><HR>\n";
-    ap_bputs(site, con->client);
-    if (f2 != NULL)
-	ap_bputs(site, f2);
-    total_bytes_sent += strlen(site);
 
-    sig = ap_psignature("", r);
-    ap_bputs(sig, con->client);
-    if (f2 != NULL)
-	ap_bputs(sig, f2);
-    total_bytes_sent += strlen(sig);
+    total_bytes_sent += ap_proxy_bputs2("</PRE><HR>\n", con->client, c);
+    total_bytes_sent += ap_proxy_bputs2(ap_psignature("", r), con->client, c);
+    total_bytes_sent += ap_proxy_bputs2("</BODY></HTML>\n", con->client, c);
 
-    site = "</BODY></HTML>\n";
-    ap_bputs(site, con->client);
-    if (f2 != NULL)
-	ap_bputs(site, f2);
-    total_bytes_sent += strlen(site);
     ap_bflush(con->client);
 
     return total_bytes_sent;
 }
 
+/* Common routine for failed authorization (i.e., missing or wrong password)
+ * to an ftp service. This causes most browsers to retry the request
+ * with username and password (which was presumably queried from the user)
+ * supplied in the Authorization: header.
+ * Note that we "invent" a realm name which consists of the
+ * ftp://user@host part of the reqest (sans password -if supplied but invalid-)
+ */
+static int ftp_unauthorized (request_rec *r, int log_it)
+{
+    r->proxyreq = 0;
+    /* Log failed requests if they supplied a password
+     * (log username/password guessing attempts)
+     */
+    if (log_it)
+	ap_log_rerror(APLOG_MARK, APLOG_INFO|APLOG_NOERRNO, r,
+		      "proxy: missing or failed auth to %s",
+		      ap_unparse_uri_components(r->pool,
+		      &r->parsed_uri, UNP_OMITPATHINFO));
+
+    ap_table_setn(r->err_headers_out, "WWW-Authenticate",
+                  ap_pstrcat(r->pool, "Basic realm=\"",
+		  ap_unparse_uri_components(r->pool, &r->parsed_uri,
+					    UNP_OMITPASSWORD|UNP_OMITPATHINFO),
+		  "\"", NULL));
+
+    return HTTP_UNAUTHORIZED;
+}
+
 /*
  * Handles direct access of ftp:// URLs
  * Original (Non-PASV) version from
  * Troy Morrison <spiffnet@zoom.com>
  * PASV added by Chuck
  */
-int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
+int ap_proxy_ftp_handler(request_rec *r, cache_req *c, char *url)
 {
-    char *host, *path, *strp, *user, *password, *parms;
+    char *host, *path, *strp, *parms;
+    char *cwd = NULL;
+    char *user = NULL;
+/*    char *account = NULL; how to supply an account in a URL? */
+    const char *password = NULL;
     const char *err;
-    int port, userlen, i, j, len, sock, dsock, rc, nocache;
-    int passlen = 0;
+    int port, i, j, len, sock, dsock, rc, nocache = 0;
     int csd = 0;
     struct sockaddr_in server;
     struct hostent server_hp;
-    struct hdr_entry *hdr;
     struct in_addr destaddr;
-    array_header *resp_hdrs;
-    BUFF *f, *cache;
+    table *resp_hdrs;
+    BUFF *f;
     BUFF *data = NULL;
     pool *p = r->pool;
     int one = 1;
     const long int zero = 0L;
     NET_SIZE_T clen;
+    struct tbl_do_args tdo;
 
     void *sconf = r->server->module_config;
     proxy_server_conf *conf =
@@ -492,59 +476,56 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
     char *pstr;
 
 /* stuff for responses */
-    char *resp;
-    int resplen;
+    char resp[MAX_STRING_LEN];
     char *size = NULL;
 
 /* we only support GET and HEAD */
 
     if (r->method_number != M_GET)
-	return NOT_IMPLEMENTED;
-
-/* allocate a buffer for the response message */
-	resplen = MAX_STRING_LEN;
-	resp = (char *)ap_palloc(r->pool, resplen);
+	return HTTP_NOT_IMPLEMENTED;
 
 /* We break the URL into host, port, path-search */
 
-    host = ap_pstrdup(p, url + 6);
-    port = DEFAULT_FTP_PORT;
-    path = strchr(host, '/');
-    if (path == NULL)
-	path = "";
-    else
-	*(path++) = '\0';
+    host = r->parsed_uri.hostname;
+    port = (r->parsed_uri.port != 0)
+	    ? r->parsed_uri.port
+	    : ap_default_port_for_request(r);
+    path = ap_pstrdup(p, r->parsed_uri.path);
+    path = (path != NULL && path[0] != '\0') ? &path[1] : "";
 
-    user = password = NULL;
-    nocache = 0;
-    strp = strchr(host, '@');
-    if (strp != NULL) {
-	(*strp++) = '\0';
-	user = host;
-	host = strp;
-/* find password */
-	strp = strchr(user, ':');
-	if (strp != NULL) {
-	    *(strp++) = '\0';
-	    password = strp;
-	    passlen = decodeenc(password);
+    /* The "Authorization:" header must be checked first.
+     * We allow the user to "override" the URL-coded user [ & password ]
+     * in the Browsers' User&Password Dialog.
+     * NOTE that this is only marginally more secure than having the
+     * password travel in plain as part of the URL, because Basic Auth
+     * simply uuencodes the plain text password. 
+     * But chances are still smaller that the URL is logged regularly.
+     */
+    if ((password = ap_table_get(r->headers_in, "Authorization")) != NULL
+	&& strcasecmp(ap_getword(r->pool, &password, ' '), "Basic") == 0
+	&& (password = ap_uudecode(r->pool, password))[0] != ':') {
+	/* Note that this allocation has to be made from r->connection->pool
+	 * because it has the lifetime of the connection.  The other allocations
+	 * are temporary and can be tossed away any time.
+	 */
+	user = ap_getword_nulls (r->connection->pool, &password, ':');
+	r->connection->ap_auth_type = "Basic";
+	r->connection->user = r->parsed_uri.user = user;
+	nocache = 1;	/* This resource only accessible with username/password */
+    }
+    else if ((user = r->parsed_uri.user) != NULL) {
+	user = ap_pstrdup(p, user);
+	decodeenc(user);
+	if ((password = r->parsed_uri.password) != NULL) {
+	    char *tmp = ap_pstrdup(p, password);
+	    decodeenc(tmp);
+	    password = tmp;
 	}
-	userlen = decodeenc(user);
-	nocache = 1;		/* don't cache when a username is supplied */
+	nocache = 1;	/* This resource only accessible with username/password */
     }
     else {
 	user = "anonymous";
-	userlen = 9;
-
 	password = "apache_proxy@";
-	passlen = strlen(password);
-    }
-
-    strp = strchr(host, ':');
-    if (strp != NULL) {
-	*(strp++) = '\0';
-	if (isdigit(*strp))
-	    port = atoi(strp);
     }
 
 /* check if ProxyBlock directive on this host */
@@ -552,7 +533,8 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
     for (i = 0; i < conf->noproxies->nelts; i++) {
 	if ((npent[i].name != NULL && strstr(host, npent[i].name) != NULL)
 	    || destaddr.s_addr == npent[i].addr.s_addr || npent[i].name[0] == '*')
-	    return ap_proxyerror(r, "Connect to remote machine blocked");
+	    return ap_proxyerror(r, HTTP_FORBIDDEN,
+				 "Connect to remote machine blocked");
     }
 
     Explain2("FTP: connect to %s:%d", host, port);
@@ -566,31 +548,30 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
     server.sin_port = htons(port);
     err = ap_proxy_host2addr(host, &server_hp);
     if (err != NULL)
-	return ap_proxyerror(r, err);	/* give up */
+	return ap_proxyerror(r, HTTP_INTERNAL_SERVER_ERROR, err);
 
     sock = ap_psocket(p, PF_INET, SOCK_STREAM, IPPROTO_TCP);
     if (sock == -1) {
-	ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 		     "proxy: error creating socket");
-	return SERVER_ERROR;
+	return HTTP_INTERNAL_SERVER_ERROR;
     }
 
-    if (conf->recv_buffer_size) {
-	if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF,
+    if (conf->recv_buffer_size > 0
+	&& setsockopt(sock, SOL_SOCKET, SO_RCVBUF,
 		       (const char *) &conf->recv_buffer_size, sizeof(int))
 	    == -1) {
-	    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 			 "setsockopt(SO_RCVBUF): Failed to set ProxyReceiveBufferSize, using default");
-	}
     }
 
     if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *) &one,
 		   sizeof(one)) == -1) {
 #ifndef _OSD_POSIX /* BS2000 has this option "always on" */
-	ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 		     "proxy: error setting reuseaddr option: setsockopt(SO_REUSEADDR)");
 	ap_pclosesocket(p, sock);
-	return SERVER_ERROR;
+	return HTTP_INTERNAL_SERVER_ERROR;
 #endif /*_OSD_POSIX*/
     }
 
@@ -618,7 +599,7 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
 #endif
     if (i == -1) {
 	ap_pclosesocket(p, sock);
-	return ap_proxyerror(r, /*HTTP_BAD_GATEWAY*/ ap_pstrcat(r->pool,
+	return ap_proxyerror(r, HTTP_BAD_GATEWAY, ap_pstrcat(r->pool,
 				"Could not connect to remote machine: ",
 				strerror(errno), NULL));
     }
@@ -631,75 +612,112 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
     ap_bsetflag(f, B_ASCII2EBCDIC|B_EBCDIC2ASCII, 1);
 #endif /*CHARSET_EBCDIC*/
 
-/* possible results: 120, 220, 421 */
+/* possible results: */
+    /* 120 Service ready in nnn minutes. */
+    /* 220 Service ready for new user. */
+    /* 421 Service not available, closing control connection. */
     ap_hard_timeout("proxy ftp", r);
-    i = ftp_getrc(f);
+    i = ftp_getrc_msg(f, resp, sizeof resp);
     Explain1("FTP: returned status %d", i);
     if (i == -1) {
 	ap_kill_timeout(r);
-	return ap_proxyerror(r, "Error reading from remote server");
+	return ap_proxyerror(r, HTTP_BAD_GATEWAY,
+			     "Error reading from remote server");
     }
+#if 0
+    if (i == 120) {
+	ap_kill_timeout(r);
+	/* RFC2068 states:
+	 * 14.38 Retry-After
+	 * 
+	 *  The Retry-After response-header field can be used with a 503 (Service
+	 *  Unavailable) response to indicate how long the service is expected to
+	 *  be unavailable to the requesting client. The value of this field can
+	 *  be either an HTTP-date or an integer number of seconds (in decimal)
+	 *  after the time of the response.
+	 *     Retry-After  = "Retry-After" ":" ( HTTP-date | delta-seconds )
+	 */
+	ap_set_header("Retry-After", ap_psprintf(p, "%u", 60*wait_mins);
+	return ap_proxyerror(r, HTTP_SERVICE_UNAVAILABLE, resp);
+    }
+#endif
     if (i != 220) {
 	ap_kill_timeout(r);
-	return BAD_GATEWAY;
+	return ap_proxyerror(r, HTTP_BAD_GATEWAY, resp);
     }
 
     Explain0("FTP: connected.");
 
-    ap_bputs("USER ", f);
-    ap_bwrite(f, user, userlen);
-    ap_bputs(CRLF, f);
+    ap_bvputs(f, "USER ", user, CRLF, NULL);
     ap_bflush(f);			/* capture any errors */
     Explain1("FTP: USER %s", user);
 
 /* possible results; 230, 331, 332, 421, 500, 501, 530 */
 /* states: 1 - error, 2 - success; 3 - send password, 4,5 fail */
+    /* 230 User logged in, proceed. */
+    /* 331 User name okay, need password. */
+    /* 332 Need account for login. */
+    /* 421 Service not available, closing control connection. */
+    /* 500 Syntax error, command unrecognized. */
+    /*     (This may include errors such as command line too long.) */
+    /* 501 Syntax error in parameters or arguments. */
+    /* 530 Not logged in. */
     i = ftp_getrc(f);
     Explain1("FTP: returned status %d", i);
     if (i == -1) {
 	ap_kill_timeout(r);
-	return ap_proxyerror(r, "Error sending to remote server");
+	return ap_proxyerror(r, HTTP_BAD_GATEWAY,
+			     "Error reading from remote server");
     }
     if (i == 530) {
 	ap_kill_timeout(r);
-	return ap_proxyerror(r, "Not logged in");
+	return ftp_unauthorized (r, 1);	/* log it: user name guessing attempt? */
     }
     if (i != 230 && i != 331) {
 	ap_kill_timeout(r);
-	return BAD_GATEWAY;
+	return HTTP_BAD_GATEWAY;
     }
 
     if (i == 331) {		/* send password */
-	if (password == NULL)
-	    return FORBIDDEN;
-	ap_bputs("PASS ", f);
-	ap_bwrite(f, password, passlen);
-	ap_bputs(CRLF, f);
+	if (password == NULL) {
+	    return ftp_unauthorized (r, 0);
+	}
+	ap_bvputs(f, "PASS ", password, CRLF, NULL);
 	ap_bflush(f);
 	Explain1("FTP: PASS %s", password);
 /* possible results 202, 230, 332, 421, 500, 501, 503, 530 */
+    /* 230 User logged in, proceed. */
+    /* 332 Need account for login. */
+    /* 421 Service not available, closing control connection. */
+    /* 500 Syntax error, command unrecognized. */
+    /* 501 Syntax error in parameters or arguments. */
+    /* 503 Bad sequence of commands. */
+    /* 530 Not logged in. */
 	i = ftp_getrc(f);
 	Explain1("FTP: returned status %d", i);
 	if (i == -1) {
 	    ap_kill_timeout(r);
-	    return ap_proxyerror(r, "Error sending to remote server");
+	    return ap_proxyerror(r, HTTP_BAD_GATEWAY,
+				 "Error reading from remote server");
 	}
 	if (i == 332) {
 	    ap_kill_timeout(r);
-	    return ap_proxyerror(r, "Need account for login");
+	    return ap_proxyerror(r, HTTP_UNAUTHORIZED,
+				 "Need account for login");
 	}
+	/* @@@ questionable -- we might as well return a 403 Forbidden here */
 	if (i == 530) {
 	    ap_kill_timeout(r);
-	    return ap_proxyerror(r, "Not logged in");
+	    return ftp_unauthorized (r, 1); /* log it: passwd guessing attempt? */
 	}
 	if (i != 230 && i != 202) {
 	    ap_kill_timeout(r);
-	    return BAD_GATEWAY;
+	    return HTTP_BAD_GATEWAY;
 	}
     }
 
-/* set the directory */
-/* this is what we must do if we don't know the OS type of the remote
+/* set the directory (walk directory component by component):
+ * this is what we must do if we don't know the OS type of the remote
  * machine
  */
     for (;;) {
@@ -709,26 +727,32 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
 	*strp = '\0';
 
 	len = decodeenc(path);
-	ap_bputs("CWD ", f);
-	ap_bwrite(f, path, len);
-	ap_bputs(CRLF, f);
+	ap_bvputs(f, "CWD ", path, CRLF, NULL);
 	ap_bflush(f);
 	Explain1("FTP: CWD %s", path);
+	*strp = '/';
 /* responses: 250, 421, 500, 501, 502, 530, 550 */
-/* 1,3 error, 2 success, 4,5 failure */
+    /* 250 Requested file action okay, completed. */
+    /* 421 Service not available, closing control connection. */
+    /* 500 Syntax error, command unrecognized. */
+    /* 501 Syntax error in parameters or arguments. */
+    /* 502 Command not implemented. */
+    /* 530 Not logged in. */
+    /* 550 Requested action not taken. */
 	i = ftp_getrc(f);
 	Explain1("FTP: returned status %d", i);
 	if (i == -1) {
 	    ap_kill_timeout(r);
-	    return ap_proxyerror(r, "Error sending to remote server");
+	    return ap_proxyerror(r, HTTP_BAD_GATEWAY,
+				 "Error reading from remote server");
 	}
 	if (i == 550) {
 	    ap_kill_timeout(r);
-	    return NOT_FOUND;
+	    return HTTP_NOT_FOUND;
 	}
 	if (i != 250) {
 	    ap_kill_timeout(r);
-	    return BAD_GATEWAY;
+	    return HTTP_BAD_GATEWAY;
 	}
 
 	path = strp + 1;
@@ -753,15 +777,22 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
 	ap_bflush(f);
 	Explain0("FTP: TYPE I");
 /* responses: 200, 421, 500, 501, 504, 530 */
+    /* 200 Command okay. */
+    /* 421 Service not available, closing control connection. */
+    /* 500 Syntax error, command unrecognized. */
+    /* 501 Syntax error in parameters or arguments. */
+    /* 504 Command not implemented for that parameter. */
+    /* 530 Not logged in. */
 	i = ftp_getrc(f);
 	Explain1("FTP: returned status %d", i);
 	if (i == -1) {
 	    ap_kill_timeout(r);
-	    return ap_proxyerror(r, "Error sending to remote server");
+	    return ap_proxyerror(r, HTTP_BAD_GATEWAY,
+				 "Error reading from remote server");
 	}
 	if (i != 200 && i != 504) {
 	    ap_kill_timeout(r);
-	    return BAD_GATEWAY;
+	    return HTTP_BAD_GATEWAY;
 	}
 /* Allow not implemented */
 	if (i == 504)
@@ -771,17 +802,17 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
 /* try to set up PASV data connection first */
     dsock = ap_psocket(p, PF_INET, SOCK_STREAM, IPPROTO_TCP);
     if (dsock == -1) {
-	ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 		     "proxy: error creating PASV socket");
 	ap_bclose(f);
 	ap_kill_timeout(r);
-	return SERVER_ERROR;
+	return HTTP_INTERNAL_SERVER_ERROR;
     }
 
     if (conf->recv_buffer_size) {
 	if (setsockopt(dsock, SOL_SOCKET, SO_RCVBUF,
 	       (const char *) &conf->recv_buffer_size, sizeof(int)) == -1) {
-	    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 			 "setsockopt(SO_RCVBUF): Failed to set ProxyReceiveBufferSize, using default");
 	}
     }
@@ -790,15 +821,20 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
     ap_bflush(f);
     Explain0("FTP: PASV command issued");
 /* possible results: 227, 421, 500, 501, 502, 530 */
+    /* 227 Entering Passive Mode (h1,h2,h3,h4,p1,p2). */
+    /* 421 Service not available, closing control connection. */
+    /* 500 Syntax error, command unrecognized. */
+    /* 501 Syntax error in parameters or arguments. */
+    /* 502 Command not implemented. */
+    /* 530 Not logged in. */
     i = ap_bgets(pasv, sizeof(pasv), f);
-
     if (i == -1) {
-	ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, r->server,
+	ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, r,
 		     "PASV: control connection is toast");
 	ap_pclosesocket(p, dsock);
 	ap_bclose(f);
 	ap_kill_timeout(r);
-	return SERVER_ERROR;
+	return HTTP_INTERNAL_SERVER_ERROR;
     }
     else {
 	pasv[i - 1] = '\0';
@@ -833,9 +869,10 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
 
 	    if (i == -1) {
 		ap_kill_timeout(r);
-		return ap_proxyerror(r, /*HTTP_BAD_GATEWAY*/ ap_pstrcat(r->pool,
-				"Could not connect to remote machine: ",
-				strerror(errno), NULL));
+		return ap_proxyerror(r, HTTP_BAD_GATEWAY,
+				     ap_pstrcat(r->pool,
+						"Could not connect to remote machine: ",
+						strerror(errno), NULL));
 	    }
 	    else {
 		pasvmode = 1;
@@ -848,31 +885,31 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
     if (!pasvmode) {		/* set up data connection */
 	clen = sizeof(struct sockaddr_in);
 	if (getsockname(sock, (struct sockaddr *) &server, &clen) < 0) {
-	    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 			 "proxy: error getting socket address");
 	    ap_bclose(f);
 	    ap_kill_timeout(r);
-	    return SERVER_ERROR;
+	    return HTTP_INTERNAL_SERVER_ERROR;
 	}
 
 	dsock = ap_psocket(p, PF_INET, SOCK_STREAM, IPPROTO_TCP);
 	if (dsock == -1) {
-	    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 			 "proxy: error creating socket");
 	    ap_bclose(f);
 	    ap_kill_timeout(r);
-	    return SERVER_ERROR;
+	    return HTTP_INTERNAL_SERVER_ERROR;
 	}
 
 	if (setsockopt(dsock, SOL_SOCKET, SO_REUSEADDR, (void *) &one,
 		       sizeof(one)) == -1) {
 #ifndef _OSD_POSIX /* BS2000 has this option "always on" */
-	    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 			 "proxy: error setting reuseaddr option");
 	    ap_pclosesocket(p, dsock);
 	    ap_bclose(f);
 	    ap_kill_timeout(r);
-	    return SERVER_ERROR;
+	    return HTTP_INTERNAL_SERVER_ERROR;
 #endif /*_OSD_POSIX*/
 	}
 
@@ -881,16 +918,16 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
 	    char buff[22];
 
 	    ap_snprintf(buff, sizeof(buff), "%s:%d", inet_ntoa(server.sin_addr), server.sin_port);
-	    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 			 "proxy: error binding to ftp data socket %s", buff);
 	    ap_bclose(f);
 	    ap_pclosesocket(p, dsock);
-	    return SERVER_ERROR;
+	    return HTTP_INTERNAL_SERVER_ERROR;
 	}
 	listen(dsock, 2);	/* only need a short queue */
     }
 
-/* set request */
+/* set request; "path" holds last path component */
     len = decodeenc(path);
 
     /* TM - if len == 0 then it must be a directory (you can't RETR nothing) */
@@ -899,119 +936,207 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
 	parms = "d";
     }
     else {
-	ap_bputs("SIZE ", f);
-	ap_bwrite(f, path, len);
-	ap_bputs(CRLF, f);
+	ap_bvputs(f, "SIZE ", path, CRLF, NULL);
 	ap_bflush(f);
 	Explain1("FTP: SIZE %s", path);
-	i = ftp_getrc_msg(f, resp, resplen);
+	i = ftp_getrc_msg(f, resp, sizeof resp);
 	Explain2("FTP: returned status %d with response %s", i, resp);
 	if (i != 500) {		/* Size command not recognized */
 	    if (i == 550) {	/* Not a regular file */
 		Explain0("FTP: SIZE shows this is a directory");
 		parms = "d";
-		ap_bputs("CWD ", f);
-		ap_bwrite(f, path, len);
-		ap_bputs(CRLF, f);
+		ap_bvputs(f, "CWD ", path, CRLF, NULL);
 		ap_bflush(f);
 		Explain1("FTP: CWD %s", path);
 		i = ftp_getrc(f);
+		/* possible results: 250, 421, 500, 501, 502, 530, 550 */
+		/* 250 Requested file action okay, completed. */
+		/* 421 Service not available, closing control connection. */
+		/* 500 Syntax error, command unrecognized. */
+		/* 501 Syntax error in parameters or arguments. */
+		/* 502 Command not implemented. */
+		/* 530 Not logged in. */
+		/* 550 Requested action not taken. */
 		Explain1("FTP: returned status %d", i);
 		if (i == -1) {
 		    ap_kill_timeout(r);
-		    return ap_proxyerror(r, "Error sending to remote server");
+		    return ap_proxyerror(r, HTTP_BAD_GATEWAY,
+					 "Error reading from remote server");
 		}
 		if (i == 550) {
 		    ap_kill_timeout(r);
-		    return NOT_FOUND;
+		    return HTTP_NOT_FOUND;
 		}
 		if (i != 250) {
 		    ap_kill_timeout(r);
-		    return BAD_GATEWAY;
+		    return HTTP_BAD_GATEWAY;
 		}
 		path = "";
 		len = 0;
 	    }
 	    else if (i == 213) { /* Size command ok */
-	        size = resp;
+		for (j = 0; j < sizeof resp && ap_isdigit(resp[j]); j++)
+			;
+		resp[j] = '\0';
+		if (resp[0] != '\0')
+		    size = ap_pstrdup(p, resp);
 	    }
 	}
     }
 
+#ifdef AUTODETECT_PWD
+    ap_bvputs(f, "PWD", CRLF, NULL);
+    ap_bflush(f);
+    Explain0("FTP: PWD");
+/* responses: 257, 500, 501, 502, 421, 550 */
+    /* 257 "<directory-name>" <commentary> */
+    /* 421 Service not available, closing control connection. */
+    /* 500 Syntax error, command unrecognized. */
+    /* 501 Syntax error in parameters or arguments. */
+    /* 502 Command not implemented. */
+    /* 550 Requested action not taken. */
+    i = ftp_getrc_msg(f, resp, sizeof resp);
+    Explain1("FTP: PWD returned status %d", i);
+    if (i == -1 || i == 421) {
+	ap_kill_timeout(r);
+	return ap_proxyerror(r, HTTP_BAD_GATEWAY,
+			     "Error reading from remote server");
+    }
+    if (i == 550) {
+	ap_kill_timeout(r);
+	return HTTP_NOT_FOUND;
+    }
+    if (i == 257) {
+	const char *dirp = resp;
+	cwd = ap_getword_conf(r->pool, &dirp);
+    }
+#endif /*AUTODETECT_PWD*/
+
     if (parms[0] == 'd') {
 	if (len != 0)
-	    ap_bputs("LIST ", f);
+	    ap_bvputs(f, "LIST ", path, CRLF, NULL);
 	else
-	    ap_bputs("LIST -lag", f);
+	    ap_bputs("LIST -lag" CRLF, f);
 	Explain1("FTP: LIST %s", (len == 0 ? "" : path));
     }
     else {
-	ap_bputs("RETR ", f);
+	ap_bvputs(f, "RETR ", path, CRLF, NULL);
 	Explain1("FTP: RETR %s", path);
     }
-    ap_bwrite(f, path, len);
-    ap_bputs(CRLF, f);
     ap_bflush(f);
 /* RETR: 110, 125, 150, 226, 250, 421, 425, 426, 450, 451, 500, 501, 530, 550
    NLST: 125, 150, 226, 250, 421, 425, 426, 450, 451, 500, 501, 502, 530 */
+    /* 110 Restart marker reply. */
+    /* 125 Data connection already open; transfer starting. */
+    /* 150 File status okay; about to open data connection. */
+    /* 226 Closing data connection. */
+    /* 250 Requested file action okay, completed. */
+    /* 421 Service not available, closing control connection. */
+    /* 425 Can't open data connection. */
+    /* 426 Connection closed; transfer aborted. */
+    /* 450 Requested file action not taken. */
+    /* 451 Requested action aborted. Local error in processing. */
+    /* 500 Syntax error, command unrecognized. */
+    /* 501 Syntax error in parameters or arguments. */
+    /* 530 Not logged in. */
+    /* 550 Requested action not taken. */
     rc = ftp_getrc(f);
     Explain1("FTP: returned status %d", rc);
     if (rc == -1) {
 	ap_kill_timeout(r);
-	return ap_proxyerror(r, "Error sending to remote server");
+	return ap_proxyerror(r, HTTP_BAD_GATEWAY,
+			     "Error reading from remote server");
     }
     if (rc == 550) {
 	Explain0("FTP: RETR failed, trying LIST instead");
 	parms = "d";
-	ap_bputs("CWD ", f);
-	ap_bwrite(f, path, len);
-	ap_bputs(CRLF, f);
+	ap_bvputs(f, "CWD ", path, CRLF, NULL);
 	ap_bflush(f);
 	Explain1("FTP: CWD %s", path);
+	/* possible results: 250, 421, 500, 501, 502, 530, 550 */
+	/* 250 Requested file action okay, completed. */
+	/* 421 Service not available, closing control connection. */
+	/* 500 Syntax error, command unrecognized. */
+	/* 501 Syntax error in parameters or arguments. */
+	/* 502 Command not implemented. */
+	/* 530 Not logged in. */
+	/* 550 Requested action not taken. */
 	rc = ftp_getrc(f);
 	Explain1("FTP: returned status %d", rc);
 	if (rc == -1) {
 	    ap_kill_timeout(r);
-	    return ap_proxyerror(r, "Error sending to remote server");
+	    return ap_proxyerror(r, HTTP_BAD_GATEWAY,
+				 "Error reading from remote server");
 	}
 	if (rc == 550) {
 	    ap_kill_timeout(r);
-	    return NOT_FOUND;
+	    return HTTP_NOT_FOUND;
 	}
 	if (rc != 250) {
 	    ap_kill_timeout(r);
-	    return BAD_GATEWAY;
+	    return HTTP_BAD_GATEWAY;
 	}
 
+#ifdef AUTODETECT_PWD
+	ap_bvputs(f, "PWD", CRLF, NULL);
+	ap_bflush(f);
+	Explain0("FTP: PWD");
+/* responses: 257, 500, 501, 502, 421, 550 */
+	/* 257 "<directory-name>" <commentary> */
+	/* 421 Service not available, closing control connection. */
+	/* 500 Syntax error, command unrecognized. */
+	/* 501 Syntax error in parameters or arguments. */
+	/* 502 Command not implemented. */
+	/* 550 Requested action not taken. */
+	i = ftp_getrc_msg(f, resp, sizeof resp);
+	Explain1("FTP: PWD returned status %d", i);
+	if (i == -1 || i == 421) {
+	    ap_kill_timeout(r);
+	    return ap_proxyerror(r, HTTP_BAD_GATEWAY,
+				 "Error reading from remote server");
+	}
+	if (i == 550) {
+	    ap_kill_timeout(r);
+	    return HTTP_NOT_FOUND;
+	}
+	if (i == 257) {
+	    const char *dirp = resp;
+	    cwd = ap_getword_conf(r->pool, &dirp);
+	}
+#endif /*AUTODETECT_PWD*/
+
 	ap_bputs("LIST -lag" CRLF, f);
 	ap_bflush(f);
 	Explain0("FTP: LIST -lag");
 	rc = ftp_getrc(f);
 	Explain1("FTP: returned status %d", rc);
 	if (rc == -1)
-	    return ap_proxyerror(r, "Error sending to remote server");
+	    return ap_proxyerror(r, HTTP_BAD_GATEWAY,
+				 "Error reading from remote server");
     }
     ap_kill_timeout(r);
     if (rc != 125 && rc != 150 && rc != 226 && rc != 250)
-	return BAD_GATEWAY;
+	return HTTP_BAD_GATEWAY;
 
-    r->status = 200;
+    r->status = HTTP_OK;
     r->status_line = "200 OK";
 
-    resp_hdrs = ap_make_array(p, 2, sizeof(struct hdr_entry));
+    resp_hdrs = ap_make_table(p, 2);
+    c->hdrs = resp_hdrs;
+
     if (parms[0] == 'd')
-	ap_proxy_add_header(resp_hdrs, "Content-Type", "text/html", HDR_REP);
+	ap_table_set(resp_hdrs, "Content-Type", "text/html");
     else {
 	if (r->content_type != NULL) {
-	    ap_proxy_add_header(resp_hdrs, "Content-Type", r->content_type,
-			     HDR_REP);
+	    ap_table_set(resp_hdrs, "Content-Type", r->content_type);
 	    Explain1("FTP: Content-Type set to %s", r->content_type);
 	}
 	else {
-	    ap_proxy_add_header(resp_hdrs, "Content-Type", "text/plain", HDR_REP);
+	    ap_table_set(resp_hdrs, "Content-Type", "text/plain");
 	}
 	if (parms[0] != 'a' && size != NULL) {
-	    ap_proxy_add_header(resp_hdrs, "Content-Length", size, HDR_REP);
+	    /* We "trust" the ftp server to really serve (size) bytes... */
+	    ap_table_set(resp_hdrs, "Content-Length", size);
 	    Explain1("FTP: Content-Length set to %s", size);
 	}
     }
@@ -1030,7 +1155,6 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
 	ap_bclose(f);
 	return i;
     }
-    cache = c->fp;
 
     if (!pasvmode) {		/* wait for connection */
 	ap_hard_timeout("proxy ftp data connect", r);
@@ -1039,13 +1163,14 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
 	    csd = accept(dsock, (struct sockaddr *) &server, &clen);
 	while (csd == -1 && errno == EINTR);
 	if (csd == -1) {
-	    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 			 "proxy: failed to accept data connection");
 	    ap_pclosesocket(p, dsock);
 	    ap_bclose(f);
 	    ap_kill_timeout(r);
-	    ap_proxy_cache_error(c);
-	    return BAD_GATEWAY;
+	    if (c != NULL)
+		c = ap_proxy_cache_error(c);
+	    return HTTP_BAD_GATEWAY;
 	}
 	ap_note_cleanups_for_socket(p, csd);
 	data = ap_bcreate(p, B_RDWR | B_SOCKET);
@@ -1066,45 +1191,38 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
 /* write status line */
     if (!r->assbackwards)
 	ap_rvputs(r, "HTTP/1.0 ", r->status_line, CRLF, NULL);
-    if (cache != NULL)
-	if (ap_bvputs(cache, "HTTP/1.0 ", r->status_line, CRLF,
-		   NULL) == -1)
-	    cache = ap_proxy_cache_error(c);
+    if (c != NULL && c->fp != NULL
+	&& ap_bvputs(c->fp, "HTTP/1.0 ", r->status_line, CRLF, NULL) == -1)
+	c = ap_proxy_cache_error(c);
 
 /* send headers */
-    len = resp_hdrs->nelts;
-    hdr = (struct hdr_entry *) resp_hdrs->elts;
-    for (i = 0; i < len; i++) {
-	if (hdr[i].field == NULL || hdr[i].value == NULL ||
-	    hdr[i].value[0] == '\0')
-	    continue;
-	if (!r->assbackwards)
-	    ap_rvputs(r, hdr[i].field, ": ", hdr[i].value, CRLF, NULL);
-	if (cache != NULL)
-	    if (ap_bvputs(cache, hdr[i].field, ": ", hdr[i].value, CRLF,
-		       NULL) == -1)
-		cache = ap_proxy_cache_error(c);
-    }
+    tdo.req = r;
+    tdo.cache = c;
+    ap_table_do(ap_proxy_send_hdr_line, &tdo, resp_hdrs, NULL);
 
     if (!r->assbackwards)
 	ap_rputs(CRLF, r);
-    if (cache != NULL)
-	if (ap_bputs(CRLF, cache) == -1)
-	    cache = ap_proxy_cache_error(c);
+    if (c != NULL && c->fp != NULL && ap_bputs(CRLF, c->fp) == -1)
+	c = ap_proxy_cache_error(c);
 
     ap_bsetopt(r->connection->client, BO_BYTECT, &zero);
     r->sent_bodyct = 1;
 /* send body */
     if (!r->header_only) {
-	if (parms[0] != 'd')
-	    ap_proxy_send_fb(data, r, cache, c);
-	else
-	    send_dir(data, r, cache, c, url);
+	if (parms[0] != 'd') {
+/* we need to set this for ap_proxy_send_fb()... */
+	    if (c != NULL)
+		c->cache_completion = 0;
+	    ap_proxy_send_fb(data, r, c);
+	} else
+	    send_dir(data, r, c, cwd);
 
 	if (rc == 125 || rc == 150)
 	    rc = ftp_getrc(f);
+
+	/* XXX: we checked for 125||150||226||250 above. This is redundant. */
 	if (rc != 226 && rc != 250)
-	    ap_proxy_cache_error(c);
+	    c = ap_proxy_cache_error(c);
     }
     else {
 /* abort the transfer */
@@ -1114,6 +1232,12 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
 	    ap_bclose(data);
 	Explain0("FTP: ABOR");
 /* responses: 225, 226, 421, 500, 501, 502 */
+    /* 225 Data connection open; no transfer in progress. */
+    /* 226 Closing data connection. */
+    /* 421 Service not available, closing control connection. */
+    /* 500 Syntax error, command unrecognized. */
+    /* 501 Syntax error in parameters or arguments. */
+    /* 502 Command not implemented. */
 	i = ftp_getrc(f);
 	Explain1("FTP: returned status %d", i);
     }
@@ -1126,6 +1250,10 @@ int ap_proxy_ftp_handler(request_rec *r, struct cache_req *c, char *url)
     ap_bflush(f);
     Explain0("FTP: QUIT");
 /* responses: 221, 500 */
+    /* 221 Service closing control connection. */
+    /* 500 Syntax error, command unrecognized. */
+    i = ftp_getrc(f);
+    Explain1("FTP: QUIT: status %d", i);
 
     if (pasvmode)
 	ap_bclose(data);
