@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1996-1998 The Apache Group.  All rights reserved.
+ * Copyright (c) 1996-1999 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -111,7 +111,7 @@ int ap_proxy_http_canon(request_rec *r, char *url, const char *scheme, int def_p
     return OK;
 }
  
-static char *proxy_location_reverse_map(request_rec *r, char *url)
+static const char *proxy_location_reverse_map(request_rec *r, const char *url)
 {
     void *sconf;
     proxy_server_conf *conf;
@@ -134,10 +134,10 @@ static char *proxy_location_reverse_map(request_rec *r, char *url)
 }
 
 /* Clear all connection-based headers from the incoming headers table */
-static void clear_connection(table *headers)
+static void clear_connection(pool *p, table *headers)
 {
-    char *name;
-    char *next = ap_table_get(headers, "Connection");
+    const char *name;
+    char *next = ap_pstrdup(p, ap_table_get(headers, "Connection"));
 
     ap_table_unset(headers, "Proxy-Connection");
     if (!next)
@@ -145,9 +145,9 @@ static void clear_connection(table *headers)
 
     while (*next) {
 	name = next;
-	while (*next && !isspace(*next) && (*next != ','))
+	while (*next && !ap_isspace(*next) && (*next != ','))
 	    ++next;
-	while (*next && (isspace(*next) || (*next == ','))) {
+	while (*next && (ap_isspace(*next) || (*next == ','))) {
 	    *next = '\0';
 	    ++next;
 	}
@@ -165,25 +165,29 @@ static void clear_connection(table *headers)
  * we return DECLINED so that we can try another proxy. (Or the direct
  * route.)
  */
-int ap_proxy_http_handler(request_rec *r, struct cache_req *c, char *url,
+int ap_proxy_http_handler(request_rec *r, cache_req *c, char *url,
 		       const char *proxyhost, int proxyport)
 {
-    char *strp;
+    const char *strp;
+    char *strp2;
     const char *err, *desthost;
     int i, j, sock, len, backasswards;
-    array_header *reqhdrs_arr, *resp_hdrs;
+    array_header *reqhdrs_arr;
+    table *resp_hdrs;
     table_entry *reqhdrs;
     struct sockaddr_in server;
     struct in_addr destaddr;
     struct hostent server_hp;
-    BUFF *f, *cache;
-    struct hdr_entry *hdr;
+    BUFF *f;
     char buffer[HUGE_STRING_LEN];
+    char portstr[32];
     pool *p = r->pool;
     const long int zero = 0L;
     int destport = 0;
     char *destportstr = NULL;
-    char *urlptr = NULL;
+    const char *urlptr = NULL;
+    const char *datestr;
+    struct tbl_do_args tdo;
 
     void *sconf = r->server->module_config;
     proxy_server_conf *conf =
@@ -215,12 +219,12 @@ int ap_proxy_http_handler(request_rec *r, struct cache_req *c, char *url,
 	desthost = q;
     }
 
-    strp = strchr(desthost, ':');
-    if (strp != NULL) {
-	*(strp++) = '\0';
-	if (isdigit(*strp)) {
-	    destport = atoi(strp);
-	    destportstr = strp;
+    strp2 = strchr(desthost, ':');
+    if (strp2 != NULL) {
+	*(strp2++) = '\0';
+	if (ap_isdigit(*strp2)) {
+	    destport = atoi(strp2);
+	    destportstr = strp2;
 	}
     }
 
@@ -229,7 +233,8 @@ int ap_proxy_http_handler(request_rec *r, struct cache_req *c, char *url,
     for (i = 0; i < conf->noproxies->nelts; i++) {
 	if ((npent[i].name != NULL && strstr(desthost, npent[i].name) != NULL)
 	    || destaddr.s_addr == npent[i].addr.s_addr || npent[i].name[0] == '*')
-	    return ap_proxyerror(r, "Connect to remote machine blocked");
+	    return ap_proxyerror(r, HTTP_FORBIDDEN,
+				 "Connect to remote machine blocked");
     }
 
     if (proxyhost != NULL) {
@@ -242,12 +247,12 @@ int ap_proxy_http_handler(request_rec *r, struct cache_req *c, char *url,
 	server.sin_port = htons(destport);
 	err = ap_proxy_host2addr(desthost, &server_hp);
 	if (err != NULL)
-	    return ap_proxyerror(r, err);	/* give up */
+	    return ap_proxyerror(r, HTTP_INTERNAL_SERVER_ERROR, err);
     }
 
     sock = ap_psocket(p, PF_INET, SOCK_STREAM, IPPROTO_TCP);
     if (sock == -1) {
-	ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 		    "proxy: error creating socket");
 	return HTTP_INTERNAL_SERVER_ERROR;
     }
@@ -256,7 +261,7 @@ int ap_proxy_http_handler(request_rec *r, struct cache_req *c, char *url,
 	if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF,
 		       (const char *) &conf->recv_buffer_size, sizeof(int))
 	    == -1) {
-	    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 			 "setsockopt(SO_RCVBUF): Failed to set ProxyReceiveBufferSize, using default");
 	}
     }
@@ -287,12 +292,12 @@ int ap_proxy_http_handler(request_rec *r, struct cache_req *c, char *url,
 	if (proxyhost != NULL)
 	    return DECLINED;	/* try again another way */
 	else
-	    return ap_proxyerror(r, /*HTTP_BAD_GATEWAY*/ ap_pstrcat(r->pool,
+	    return ap_proxyerror(r, HTTP_BAD_GATEWAY, ap_pstrcat(r->pool,
 				"Could not connect to remote machine: ",
 				strerror(errno), NULL));
     }
 
-    clear_connection(r->headers_in);	/* Strip connection-based headers */
+    clear_connection(r->pool, r->headers_in);	/* Strip connection-based headers */
 
     f = ap_bcreate(p, B_RDWR | B_SOCKET);
     ap_bpushfd(f, sock, sock);
@@ -300,28 +305,57 @@ int ap_proxy_http_handler(request_rec *r, struct cache_req *c, char *url,
     ap_hard_timeout("proxy send", r);
     ap_bvputs(f, r->method, " ", proxyhost ? url : urlptr, " HTTP/1.0" CRLF,
 	   NULL);
-    ap_bvputs(f, "Host: ", desthost, NULL);
     if (destportstr != NULL && destport != DEFAULT_HTTP_PORT)
-	ap_bvputs(f, ":", destportstr, CRLF, NULL);
+	ap_bvputs(f, "Host: ", desthost, ":", destportstr, CRLF, NULL);
     else
-	ap_bputs(CRLF, f);
+	ap_bvputs(f, "Host: ", desthost, CRLF, NULL);
 
-    reqhdrs_arr = table_elts(r->headers_in);
+    if (conf->viaopt == via_block) {
+	/* Block all outgoing Via: headers */
+	ap_table_unset(r->headers_in, "Via");
+    } else if (conf->viaopt != via_off) {
+	/* Create a "Via:" request header entry and merge it */
+	i = ap_get_server_port(r);
+	if (ap_is_default_port(i,r)) {
+	    strcpy(portstr,"");
+	} else {
+	    ap_snprintf(portstr, sizeof portstr, ":%d", i);
+	}
+	/* Generate outgoing Via: header with/without server comment: */
+	ap_table_mergen(r->headers_in, "Via",
+		    (conf->viaopt == via_full)
+			? ap_psprintf(p, "%d.%d %s%s (%s)",
+				HTTP_VERSION_MAJOR(r->proto_num),
+				HTTP_VERSION_MINOR(r->proto_num),
+				ap_get_server_name(r), portstr,
+				SERVER_BASEVERSION)
+			: ap_psprintf(p, "%d.%d %s%s",
+				HTTP_VERSION_MAJOR(r->proto_num),
+				HTTP_VERSION_MINOR(r->proto_num),
+				ap_get_server_name(r), portstr)
+			);
+    }
+
+    reqhdrs_arr = ap_table_elts(r->headers_in);
     reqhdrs = (table_entry *) reqhdrs_arr->elts;
     for (i = 0; i < reqhdrs_arr->nelts; i++) {
 	if (reqhdrs[i].key == NULL || reqhdrs[i].val == NULL
 	/* Clear out headers not to send */
 	    || !strcasecmp(reqhdrs[i].key, "Host")	/* Already sent */
-	    ||!strcasecmp(reqhdrs[i].key, "Proxy-Authorization"))
+	    /* XXX: @@@ FIXME: "Proxy-Authorization" should *only* be 
+	     * suppressed if THIS server requested the authentication,
+	     * not when a frontend proxy requested it!
+	     */
+	    || !strcasecmp(reqhdrs[i].key, "Proxy-Authorization"))
 	    continue;
 	ap_bvputs(f, reqhdrs[i].key, ": ", reqhdrs[i].val, CRLF, NULL);
     }
 
     ap_bputs(CRLF, f);
-/* send the request data, if any. N.B. should we trap SIGPIPE ? */
+/* send the request data, if any. */
 
     if (ap_should_client_block(r)) {
-	while ((i = ap_get_client_block(r, buffer, HUGE_STRING_LEN)) > 0)
+	while ((i = ap_get_client_block(r, buffer, sizeof buffer)) > 0)
 	    ap_bwrite(f, buffer, i);
     }
     ap_bflush(f);
@@ -329,20 +363,30 @@ int ap_proxy_http_handler(request_rec *r, struct cache_req *c, char *url,
 
     ap_hard_timeout("proxy receive", r);
 
-    len = ap_bgets(buffer, HUGE_STRING_LEN - 1, f);
+    len = ap_bgets(buffer, sizeof buffer - 1, f);
     if (len == -1 || len == 0) {
 	ap_bclose(f);
 	ap_kill_timeout(r);
-	return ap_proxyerror(r, "Error reading from remote server");
+	ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
+		     "ap_bgets() - proxy receive - Error reading from remote server %s",
+		     proxyhost ? proxyhost : desthost);
+	return ap_proxyerror(r, HTTP_BAD_GATEWAY,
+			     "Error reading from remote server");
     }
 
 /* Is it an HTTP/1 response?  This is buggy if we ever see an HTTP/1.10 */
     if (ap_checkmask(buffer, "HTTP/#.# ###*")) {
-/* If not an HTTP/1 messsage or if the status line was > 8192 bytes */
+	int major, minor;
+	if (2 != sscanf(buffer, "HTTP/%u.%u", &major, &minor)) {
+	    major = 1;
+	    minor = 0;
+	}
+
+/* If not an HTTP/1 message or if the status line was > 8192 bytes */
 	if (buffer[5] != '1' || buffer[len - 1] != '\n') {
 	    ap_bclose(f);
 	    ap_kill_timeout(r);
-	    return BAD_GATEWAY;
+	    return HTTP_BAD_GATEWAY;
 	}
 	backasswards = 0;
 	buffer[--len] = '\0';
@@ -356,9 +400,36 @@ int ap_proxy_http_handler(request_rec *r, struct cache_req *c, char *url,
 /* N.B. for HTTP/1.0 clients, we have to fold line-wrapped headers */
 /* Also, take care with headers with multiple occurences. */
 
-	resp_hdrs = ap_proxy_read_headers(p, buffer, HUGE_STRING_LEN, f);
+	resp_hdrs = ap_proxy_read_headers(r, buffer, HUGE_STRING_LEN, f);
+	if (resp_hdrs == NULL) {
+	    ap_log_error(APLOG_MARK, APLOG_WARNING|APLOG_NOERRNO, r->server,
+		 "proxy: Bad HTTP/%d.%d header returned by %s (%s)",
+		 major, minor, r->uri, r->method);
+	    resp_hdrs = ap_make_table(p, 20);
+	    nocache = 1;    /* do not cache this broken file */
+	}
 
-	clear_connection((table *) resp_hdrs);	/* Strip Connection hdrs */
+	if (conf->viaopt != via_off && conf->viaopt != via_block) {
+	    /* Create a "Via:" response header entry and merge it */
+	    i = ap_get_server_port(r);
+	    if (ap_is_default_port(i,r)) {
+		strcpy(portstr,"");
+	    } else {
+		ap_snprintf(portstr, sizeof portstr, ":%d", i);
+	    }
+	    ap_table_mergen((table *)resp_hdrs, "Via",
+			    (conf->viaopt == via_full)
+			    ? ap_psprintf(p, "%d.%d %s%s (%s)",
+				major, minor,
+				ap_get_server_name(r), portstr,
+				SERVER_BASEVERSION)
+			    : ap_psprintf(p, "%d.%d %s%s",
+				major, minor,
+				ap_get_server_name(r), portstr)
+			    );
+	}
+
+	clear_connection(p, resp_hdrs);	/* Strip Connection hdrs */
     }
     else {
 /* an http/0.9 response */
@@ -367,29 +438,28 @@ int ap_proxy_http_handler(request_rec *r, struct cache_req *c, char *url,
 	r->status_line = "200 OK";
 
 /* no headers */
-	resp_hdrs = ap_make_array(p, 2, sizeof(struct hdr_entry));
+	resp_hdrs = ap_make_table(p, 20);
     }
 
+    c->hdrs = resp_hdrs;
+
     ap_kill_timeout(r);
 
 /*
  * HTTP/1.0 requires us to accept 3 types of dates, but only generate
  * one type
  */
+    if ((datestr = ap_table_get(resp_hdrs, "Date")) != NULL)
+	ap_table_set(resp_hdrs, "Date", ap_proxy_date_canon(p, datestr));
+    if ((datestr = ap_table_get(resp_hdrs, "Last-Modified")) != NULL)
+	ap_table_set(resp_hdrs, "Last-Modified", ap_proxy_date_canon(p, datestr));
+    if ((datestr = ap_table_get(resp_hdrs, "Expires")) != NULL)
+	ap_table_set(resp_hdrs, "Expires", ap_proxy_date_canon(p, datestr));
 
-    hdr = (struct hdr_entry *) resp_hdrs->elts;
-    for (i = 0; i < resp_hdrs->nelts; i++) {
-	if (hdr[i].value[0] == '\0')
-	    continue;
-	strp = hdr[i].field;
-	if (strcasecmp(strp, "Date") == 0 ||
-	    strcasecmp(strp, "Last-Modified") == 0 ||
-	    strcasecmp(strp, "Expires") == 0)
-	    hdr[i].value = ap_proxy_date_canon(p, hdr[i].value);
-	if (strcasecmp(strp, "Location") == 0 ||
-	    strcasecmp(strp, "URI") == 0)
-	    hdr[i].value = proxy_location_reverse_map(r, hdr[i].value);
-    }
+    if ((datestr = ap_table_get(resp_hdrs, "Location")) != NULL)
+	ap_table_set(resp_hdrs, "Location", proxy_location_reverse_map(r, datestr));
+    if ((datestr = ap_table_get(resp_hdrs, "URI")) != NULL)
+	ap_table_set(resp_hdrs, "URI", proxy_location_reverse_map(r, datestr));
 
 /* check if NoCache directive on this host */
     for (i = 0; i < conf->nocaches->nelts; i++) {
@@ -404,46 +474,32 @@ int ap_proxy_http_handler(request_rec *r, struct cache_req *c, char *url,
 	return i;
     }
 
-    cache = c->fp;
-
     ap_hard_timeout("proxy receive", r);
 
 /* write status line */
     if (!r->assbackwards)
 	ap_rvputs(r, "HTTP/1.0 ", r->status_line, CRLF, NULL);
-    if (cache != NULL)
-	if (ap_bvputs(cache, "HTTP/1.0 ", r->status_line, CRLF, NULL) == -1)
-	    cache = ap_proxy_cache_error(c);
+    if (c != NULL && c->fp != NULL &&
+	ap_bvputs(c->fp, "HTTP/1.0 ", r->status_line, CRLF, NULL) == -1)
+	c = ap_proxy_cache_error(c);
 
 /* send headers */
-    for (i = 0; i < resp_hdrs->nelts; i++) {
-	if (hdr[i].field == NULL || hdr[i].value == NULL ||
-	    hdr[i].value[0] == '\0')
-	    continue;
-	if (!r->assbackwards) {
-	    ap_rvputs(r, hdr[i].field, ": ", hdr[i].value, CRLF, NULL);
-	    ap_table_set(r->headers_out, hdr[i].field, hdr[i].value);
-	}
-	if (cache != NULL)
-	    if (ap_bvputs(cache, hdr[i].field, ": ", hdr[i].value, CRLF,
-		       NULL) == -1)
-		cache = ap_proxy_cache_error(c);
-    }
+    tdo.req = r;
+    tdo.cache = c;
+    ap_table_do(ap_proxy_send_hdr_line, &tdo, resp_hdrs, NULL);
 
     if (!r->assbackwards)
 	ap_rputs(CRLF, r);
-    if (cache != NULL)
-	if (ap_bputs(CRLF, cache) == -1)
-	    cache = ap_proxy_cache_error(c);
+    if (c != NULL && c->fp != NULL && ap_bputs(CRLF, c->fp) == -1)
+	c = ap_proxy_cache_error(c);
 
     ap_bsetopt(r->connection->client, BO_BYTECT, &zero);
     r->sent_bodyct = 1;
 /* Is it an HTTP/0.9 respose? If so, send the extra data */
     if (backasswards) {
 	ap_bwrite(r->connection->client, buffer, len);
-	if (cache != NULL)
-	    if (ap_bwrite(f, buffer, len) != len)
-		cache = ap_proxy_cache_error(c);
+	if (c != NULL && c->fp != NULL && ap_bwrite(c->fp, buffer, len) != len)
+	    c = ap_proxy_cache_error(c);
     }
     ap_kill_timeout(r);
 
@@ -458,8 +514,11 @@ int ap_proxy_http_handler(request_rec *r, struct cache_req *c, char *url,
 /* send body */
 /* if header only, then cache will be NULL */
 /* HTTP/1.0 tells us to read to EOF, rather than content-length bytes */
-    if (!r->header_only)
-	ap_proxy_send_fb(f, r, cache, c);
+    if (!r->header_only) {
+/* we need to set this for ap_proxy_send_fb()... */
+	c->cache_completion = conf->cache.cache_completion;
+	ap_proxy_send_fb(f, r, c);
+    }
 
     ap_proxy_cache_tidy(c);
 
