@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -138,7 +138,7 @@ static const char *order(cmd_parms *cmd, void *dv, char *arg)
 
 static int is_ip(const char *host)
 {
-    while ((*host == '.') || isdigit(*host))
+    while ((*host == '.') || ap_isdigit(*host))
 	host++;
     return (*host == '\0');
 }
@@ -204,7 +204,7 @@ static const char *allow_cmd(cmd_parms *cmd, void *dv, char *from, char *where)
 	a->x.ip.mask = mask;
 
     }
-    else if (isdigit(*where) && is_ip(where)) {
+    else if (ap_isdigit(*where) && is_ip(where)) {
 	/* legacy syntax for ip addrs: a.b.c. ==> a.b.c.0/24 for example */
 	int shift;
 	char *t;
@@ -218,11 +218,11 @@ static const char *allow_cmd(cmd_parms *cmd, void *dv, char *from, char *where)
 	shift = 24;
 	while (*s) {
 	    t = s;
-	    if (!isdigit(*t)) {
+	    if (!ap_isdigit(*t)) {
 		a->type = T_FAIL;
 		return "invalid ip address";
 	    }
-	    while (isdigit(*t)) {
+	    while (ap_isdigit(*t)) {
 		++t;
 	    }
 	    if (*t == '.') {
@@ -376,9 +376,9 @@ static int check_dir_access(request_rec *r)
 
     if (ret == FORBIDDEN
 	&& (ap_satisfies(r) != SATISFY_ANY || !ap_some_auth_required(r))) {
-	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
-		  "client %pI denied by server configuration: %s",
-		  &r->connection->remote_addr, r->filename);
+	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
+		  "client denied by server configuration: %s",
+		  r->filename);
     }
 
     return ret;
