@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -217,10 +217,10 @@ static int anon_authenticate_basic_user(request_rec *r)
     (anon_auth_config_rec *) ap_get_module_config(r->per_dir_config,
 					       &anon_auth_module);
     conn_rec *c = r->connection;
-    char *send_pw;
+    const char *sent_pw;
     int res = DECLINED;
 
-    if ((res = ap_get_basic_auth_pw(r, &send_pw)))
+    if ((res = ap_get_basic_auth_pw(r, &sent_pw)))
 	return res;
 
     /* Ignore if we are not configured */
@@ -246,23 +246,23 @@ static int anon_authenticate_basic_user(request_rec *r)
     /* username is OK */
 	   (res == OK)
     /* password been filled out ? */
-	   && ((!sec->auth_anon_mustemail) || strlen(send_pw))
+	   && ((!sec->auth_anon_mustemail) || strlen(sent_pw))
     /* does the password look like an email address ? */
 	   && ((!sec->auth_anon_verifyemail)
-	       || ((strpbrk("@", send_pw) != NULL)
-		   && (strpbrk(".", send_pw) != NULL)))) {
+	       || ((strpbrk("@", sent_pw) != NULL)
+		   && (strpbrk(".", sent_pw) != NULL)))) {
 	if (sec->auth_anon_logemail && ap_is_initial_req(r)) {
-	    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, r->server,
+	    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, r,
 			"Anonymous: Passwd <%s> Accepted",
-			send_pw ? send_pw : "\'none\'");
+			sent_pw ? sent_pw : "\'none\'");
 	}
 	return OK;
     }
     else {
 	if (sec->auth_anon_authoritative) {
-	    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
+	    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
 			"Anonymous: Authoritative, Passwd <%s> not accepted",
-			send_pw ? send_pw : "\'none\'");
+			sent_pw ? sent_pw : "\'none\'");
 	    return AUTH_REQUIRED;
 	}
 	/* Drop out the bottom to return DECLINED */
