@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -96,6 +96,11 @@
 #include "http_log.h"
 #include "http_protocol.h"
 #include <db.h>
+#include "ap_md5.h"
+
+#if defined(DB_VERSION_MAJOR) && (DB_VERSION_MAJOR == 2)
+#define DB2
+#endif
 
 typedef struct {
 
@@ -151,22 +156,37 @@ static char *get_db_pw(request_rec *r, char *user, const char *auth_dbpwfile)
     DBT d, q;
     char *pw = NULL;
 
+    memset(&d, 0, sizeof(d));
+    memset(&q, 0, sizeof(q));
+
     q.data = user;
     q.size = strlen(q.data);
 
+#ifdef DB2
+    if (db_open(auth_dbpwfile, DB_HASH, DB_RDONLY, 0664, NULL, NULL, &f) != 0) {
+#else
     if (!(f = dbopen(auth_dbpwfile, O_RDONLY, 0664, DB_HASH, NULL))) {
-	ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+#endif
+	ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 		    "could not open db auth file: %s", auth_dbpwfile);
 	return NULL;
     }
 
+#ifdef DB2
+    if (!((f->get) (f, NULL, &q, &d, 0))) {
+#else
     if (!((f->get) (f, &q, &d, 0))) {
+#endif
 	pw = ap_palloc(r->pool, d.size + 1);
 	strncpy(pw, d.data, d.size);
 	pw[d.size] = '\0';	/* Terminate the string */
     }
 
+#ifdef DB2
+    (f->close) (f, 0);
+#else
     (f->close) (f);
+#endif
     return pw;
 }
 
@@ -205,7 +225,9 @@ static int db_authenticate_basic_user(request_rec *r)
     (db_auth_config_rec *) ap_get_module_config(r->per_dir_config,
 					     &db_auth_module);
     conn_rec *c = r->connection;
-    char *sent_pw, *real_pw, *colon_pw;
+    const char *sent_pw;
+    char *real_pw, *colon_pw;
+    char *invalid_pw;
     int res;
 
     if ((res = ap_get_basic_auth_pw(r, &sent_pw)))
@@ -217,19 +239,21 @@ static int db_authenticate_basic_user(request_rec *r)
     if (!(real_pw = get_db_pw(r, c->user, sec->auth_dbpwfile))) {
 	if (!(sec->auth_dbauthoritative))
 	    return DECLINED;
-	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
+	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
 		    "DB user %s not found: %s", c->user, r->filename);
 	ap_note_basic_auth_failure(r);
 	return AUTH_REQUIRED;
     }
     /* Password is up to first : if exists */
     colon_pw = strchr(real_pw, ':');
-    if (colon_pw)
+    if (colon_pw) {
 	*colon_pw = '\0';
-    /* anyone know where the prototype for crypt is? */
-    if (strcmp(real_pw, (char *) crypt(sent_pw, real_pw))) {
-	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
-		    "DB user %s: password mismatch: %s", c->user, r->uri);
+    }
+    invalid_pw = ap_validate_password(sent_pw, real_pw);
+    if (invalid_pw != NULL) {
+	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
+		      "DB user %s: authentication failure for \"%s\": %s",
+		      c->user, r->uri, invalid_pw);
 	ap_note_basic_auth_failure(r);
 	return AUTH_REQUIRED;
     }
@@ -246,7 +270,7 @@ static int db_check_auth(request_rec *r)
     char *user = r->connection->user;
     int m = r->method_number;
 
-    array_header *reqs_arr = ap_requires(r);
+    const array_header *reqs_arr = ap_requires(r);
     require_line *reqs = reqs_arr ? (require_line *) reqs_arr->elts : NULL;
 
     register int x;
@@ -264,7 +288,7 @@ static int db_check_auth(request_rec *r)
 	    continue;
 
 	t = reqs[x].requirement;
-	w = ap_getword(r->pool, &t, ' ');
+	w = ap_getword_white(r->pool, &t);
 
 	if (!strcmp(w, "group") && sec->auth_dbgrpfile) {
 	    const char *orig_groups, *groups;
@@ -273,7 +297,7 @@ static int db_check_auth(request_rec *r)
 	    if (!(groups = get_db_grp(r, user, sec->auth_dbgrpfile))) {
 		if (!(sec->auth_dbauthoritative))
 		    return DECLINED;
-		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
+		ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
 			    "user %s not in DB group file %s: %s",
 			    user, sec->auth_dbgrpfile, r->filename);
 		ap_note_basic_auth_failure(r);
@@ -281,7 +305,7 @@ static int db_check_auth(request_rec *r)
 	    }
 	    orig_groups = groups;
 	    while (t[0]) {
-		w = ap_getword(r->pool, &t, ' ');
+		w = ap_getword_white(r->pool, &t);
 		groups = orig_groups;
 		while (groups[0]) {
 		    v = ap_getword(r->pool, &groups, ',');
@@ -289,7 +313,7 @@ static int db_check_auth(request_rec *r)
 			return OK;
 		}
 	    }
-	    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
+	    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
 			"user %s not in right group: %s", user, r->filename);
 	    ap_note_basic_auth_failure(r);
 	    return AUTH_REQUIRED;
