@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -75,6 +75,7 @@
 #include "http_log.h"
 #include "http_protocol.h"
 #include <ndbm.h>
+#include "ap_md5.h"
 
 /*
  * Module definition information - the part between the -START and -END
@@ -154,7 +155,7 @@ static char *get_dbm_pw(request_rec *r, char *user, char *auth_dbmpwfile)
 
 
     if (!(f = dbm_open(auth_dbmpwfile, O_RDONLY, 0664))) {
-	ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
+	ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 		    "could not open dbm auth file: %s", auth_dbmpwfile);
 	return NULL;
     }
@@ -206,7 +207,9 @@ static int dbm_authenticate_basic_user(request_rec *r)
     (dbm_auth_config_rec *) ap_get_module_config(r->per_dir_config,
 					      &dbm_auth_module);
     conn_rec *c = r->connection;
-    char *sent_pw, *real_pw, *colon_pw;
+    const char *sent_pw;
+    char *real_pw, *colon_pw;
+    char *invalid_pw;
     int res;
 
     if ((res = ap_get_basic_auth_pw(r, &sent_pw)))
@@ -218,19 +221,21 @@ static int dbm_authenticate_basic_user(request_rec *r)
     if (!(real_pw = get_dbm_pw(r, c->user, sec->auth_dbmpwfile))) {
 	if (!(sec->auth_dbmauthoritative))
 	    return DECLINED;
-	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
+	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
 		    "DBM user %s not found: %s", c->user, r->filename);
 	ap_note_basic_auth_failure(r);
 	return AUTH_REQUIRED;
     }
     /* Password is up to first : if exists */
     colon_pw = strchr(real_pw, ':');
-    if (colon_pw)
+    if (colon_pw) {
 	*colon_pw = '\0';
-    /* anyone know where the prototype for crypt is? */
-    if (strcmp(real_pw, (char *) crypt(sent_pw, real_pw))) {
-	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
-		    "user %s: password mismatch: %s", c->user, r->uri);
+    }
+    invalid_pw = ap_validate_password(sent_pw, real_pw);
+    if (invalid_pw != NULL) {
+	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
+		      "DBM user %s: authentication failure for \"%s\": %s",
+		      c->user, r->uri, invalid_pw);
 	ap_note_basic_auth_failure(r);
 	return AUTH_REQUIRED;
     }
@@ -247,7 +252,7 @@ static int dbm_check_auth(request_rec *r)
     char *user = r->connection->user;
     int m = r->method_number;
 
-    array_header *reqs_arr = ap_requires(r);
+    const array_header *reqs_arr = ap_requires(r);
     require_line *reqs = reqs_arr ? (require_line *) reqs_arr->elts : NULL;
 
     register int x;
@@ -265,7 +270,7 @@ static int dbm_check_auth(request_rec *r)
 	    continue;
 
 	t = reqs[x].requirement;
-	w = ap_getword(r->pool, &t, ' ');
+	w = ap_getword_white(r->pool, &t);
 
 	if (!strcmp(w, "group") && sec->auth_dbmgrpfile) {
 	    const char *orig_groups, *groups;
@@ -274,7 +279,7 @@ static int dbm_check_auth(request_rec *r)
 	    if (!(groups = get_dbm_grp(r, user, sec->auth_dbmgrpfile))) {
 		if (!(sec->auth_dbmauthoritative))
 		    return DECLINED;
-		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
+		ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
 			    "user %s not in DBM group file %s: %s",
 			    user, sec->auth_dbmgrpfile, r->filename);
 		ap_note_basic_auth_failure(r);
@@ -282,7 +287,7 @@ static int dbm_check_auth(request_rec *r)
 	    }
 	    orig_groups = groups;
 	    while (t[0]) {
-		w = ap_getword(r->pool, &t, ' ');
+		w = ap_getword_white(r->pool, &t);
 		groups = orig_groups;
 		while (groups[0]) {
 		    v = ap_getword(r->pool, &groups, ',');
@@ -290,7 +295,7 @@ static int dbm_check_auth(request_rec *r)
 			return OK;
 		}
 	    }
-	    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
+	    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
 			"user %s not in right group: %s",
 			user, r->filename);
 	    ap_note_basic_auth_failure(r);
