@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -145,7 +145,7 @@ static void *merge_env_server_configs(pool *p, void *basev, void *addv)
 
     new_table = ap_copy_table(p, base->vars);
 
-    arr = table_elts(add->vars);
+    arr = ap_table_elts(add->vars);
     elts = (table_entry *)arr->elts;
 
     for (i = 0; i < arr->nelts; ++i) {
