@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1996-1998 The Apache Group.  All rights reserved.
+ * Copyright (c) 1996-1999 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -127,7 +127,7 @@
      */
 #ifndef NO_DBM_REWRITEMAP
 #include <ndbm.h>
-#if (__FreeBSD__)
+#if defined(__FreeBSD__) || (defined(DB_LOCK) && defined(DB_SHMEM))
 #define NDBM_FILE_SUFFIX ".db"
 #else
 #define NDBM_FILE_SUFFIX ".pag"
@@ -202,6 +202,7 @@
 #define RULEFLAG_FORBIDDEN          1<<9
 #define RULEFLAG_GONE               1<<10
 #define RULEFLAG_QSAPPEND           1<<11
+#define RULEFLAG_NOCASE             1<<12
 
 #define MAPTYPE_TXT                 1<<0
 #define MAPTYPE_DBM                 1<<1
@@ -218,6 +219,9 @@
 #define CACHEMODE_TS                1<<0
 #define CACHEMODE_TTL               1<<1
 
+#define CACHE_TLB_ROWS 1024
+#define CACHE_TLB_COLS 4
+
 #ifndef FALSE
 #define FALSE 0
 #define TRUE  !FALSE
@@ -240,10 +244,6 @@
 
 #define MAX_NMATCH    10
 
-#define MAPFILE_PATTERN "^([^ \t]+)[ \t]+([^ \t]+).*$"
-#define MAPFILE_OUTPUT  "$1,$2"
-
-
 /*
 **
 **  our private data structures we handle with
@@ -316,17 +316,23 @@ typedef struct {
 } rewrite_perdir_conf;
 
 
-    /* the cache structures */
-
+    /* the cache structures,
+     * a 4-way hash table with LRU functionality
+     */
 typedef struct cacheentry {
     time_t time;
     char  *key;
     char  *value;
 } cacheentry;
 
+typedef struct tlbentry {
+    int t[CACHE_TLB_COLS];
+} cachetlbentry;
+
 typedef struct cachelist {
     char         *resource;
     array_header *entries;
+    array_header *tlb;
 } cachelist;
 
 typedef struct cache {
@@ -334,9 +340,10 @@ typedef struct cache {
     array_header *lists;
 } cache;
 
-    /* the regex structure for the
-       substitution of backreferences */
 
+    /* the regex structure for the
+     * substitution of backreferences
+     */
 typedef struct backrefinfo {
     char *source;
     int nsub;
@@ -427,6 +434,8 @@ static char *lookup_map_internal(request_rec *r,
                                  char *key);
 static char *rewrite_mapfunc_toupper(request_rec *r, char *key);
 static char *rewrite_mapfunc_tolower(request_rec *r, char *key);
+static char *rewrite_mapfunc_escape(request_rec *r, char *key);
+static char *rewrite_mapfunc_unescape(request_rec *r, char *key);
 static char *select_random_value_part(request_rec *r, char *value);
 static void  rewrite_rand_init(void);
 static int   rewrite_rand(int l, int h);
