@@ -10,16 +10,14 @@
  * by Alexei Kosut, based on htpasswd.c, by Rob McCool
  */
 
-#include "conf.h"
+#include "ap_config.h"
 #include <sys/types.h>
-#ifdef MPE
+#if defined(MPE) || defined(QNX)
 #include <signal.h>
 #else
 #include <sys/signal.h>
 #endif
-
-/* This is probably the easiest way to do it */
-#include "../main/md5c.c"
+#include "ap_md5.h"
 
 #define LF 10
 #define CR 13
@@ -196,7 +194,7 @@ int main(int argc, char *argv[])
     }
     fclose(f);
     fclose(tfp);
-#if defined(__EMX__) || defined(WIN32)
+#if defined(OS2) || defined(WIN32)
     sprintf(command, "copy \"%s\" \"%s\"", tn, argv[1]);
 #else
     sprintf(command, "cp %s %s", tn, argv[1]);
