@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -109,7 +109,7 @@ int initgroups(const char *name, gid_t basegid)
 #define AP_ENVBUF 256
 
 extern char **environ;
-static FILE *log;
+static FILE *log = NULL;
 
 char *safe_env_lst[] =
 {
@@ -341,6 +341,33 @@ int main(int argc, char *argv[])
 	actual_gname = strdup(target_gname);
     }
 
+#ifdef _OSD_POSIX
+    /*
+     * Initialize BS2000 user environment
+     */
+    {
+	pid_t pid;
+	int status;
+
+	switch (pid = ufork(target_uname))
+	{
+	case -1:	/* Error */
+	    log_err("failed to setup bs2000 environment for user %s: %s\n",
+		    target_uname, strerror(errno));
+	    exit(150);
+	case 0:	/* Child */
+	    break;
+	default:	/* Father */
+	    while (pid != waitpid(pid, &status, 0))
+		;
+	    /* @@@ FIXME: should we deal with STOP signals as well? */
+	    if (WIFSIGNALED(status))
+		kill (getpid(), WTERMSIG(status));
+	    exit(WEXITSTATUS(status));
+	}
+    }
+#endif /*_OSD_POSIX*/
+
     /*
      * Save these for later since initgroups will hose the struct
      */
@@ -500,15 +527,30 @@ int main(int argc, char *argv[])
     /* 
      * Be sure to close the log file so the CGI can't
      * mess with it.  If the exec fails, it will be reopened 
-     * automatically when log_err is called.
+     * automatically when log_err is called.  Note that the log
+     * might not actually be open if LOG_EXEC isn't defined.
+     * However, the "log" cell isn't ifdef'd so let's be defensive
+     * and assume someone might have done something with it
+     * outside an ifdef'd LOG_EXEC block.
      */
-    fclose(log);
-    log = NULL;
+    if (log != NULL) {
+	fclose(log);
+	log = NULL;
+    }
 
     /*
      * Execute the command, replacing our image with its own.
      */
+#ifdef NEED_HASHBANG_EMUL
+    /* We need the #! emulation when we want to execute scripts */
+    {
+	extern char **environ;
+
+	ap_execve(cmd, &argv[3], environ);
+    }
+#else /*NEED_HASHBANG_EMUL*/
     execv(cmd, &argv[3]);
+#endif /*NEED_HASHBANG_EMUL*/
 
     /*
      * (I can't help myself...sorry.)
